# Copyright 2014-2017 Insight Software Consortium.
# Copyright 2004-2009 Roman Yakovenko.
# Distributed under the Boost Software License, Version 1.0.
# See http://www.boost.org/LICENSE_1_0.txt
#
# The initial version of the directory_cache_t class was written
# by Matthias Baas (baas@ira.uka.de).

"""
directory cache implementation.

This module contains the implementation of a cache that uses individual
files, stored in a dedicated cache directory, to store the cached contents.

The :class:`parser.directory_cache_t` class instance could be passed as the
`cache` argument of the :func:`parser.parse` function.
"""

import os
import os.path
import gzip
import hashlib
try:
    import cPickle as pickle
except ImportError:
    import pickle

from . import declarations_cache


class index_entry_t(object):

    """
    Entry of the index table in the directory cache index.

    Each cached header file (i.e. each .cache file) has a corresponding
    index_entry_t object. This object is used to determine whether the
    cache file with the declarations is still valid or not.

    This class is a helper class for the directory_cache_t class.
    """

    def __init__(self, filesigs, configsig):
        """
        :param filesigs: a list of tuples( `fileid`, `sig`)...
        :param configsig: the signature of the configuration object.
        """

        self.filesigs = filesigs
        self.configsig = configsig

    def __getstate__(self):
        return self.filesigs, self.configsig

    def __setstate__(self, state):
        self.filesigs, self.configsig = state


# pylint: disable=W0622
class directory_cache_t(declarations_cache.cache_base_t):

    """cache class that stores its data as multiple files inside a directory.

    The cache stores one index file called `index.dat` which is always
    read by the cache when the cache object is created. Each header file
    will have its corresponding .cache file that stores the declarations
    found in the header file. The index file is used to determine whether
    a .cache file is still valid or not (by checking if one of the dependent
    files (i.e. the header file itself and all included files) have been
    modified since the last run).
    """

    def __init__(self, directory="cache", compression=False, sha1_sigs=True):
        """
        :param directory: cache directory path, it is created,
                          if it does not exist

        :param compression: if `True`, the cache files will be compressed
                            using `gzip`

        :param sha1_sigs: `sha1_sigs` determines whether file modifications is
                         checked by computing a `sha1` digest or by checking
                         the modification date
        """

        declarations_cache.cache_base_t.__init__(self)

        # Cache directory
        self.__dir = os.path.abspath(directory)

        # Flag that determines whether the cache files will be compressed
        self.__compression = compression

        # Flag that determines whether the signature is a sha1 digest or
        # the modification time
        # (this flag is passed to the filename_repository_t class)
        self.__sha1_sigs = sha1_sigs

        # Filename repository
        self.__filename_rep = filename_repository_t(self.__sha1_sigs)

        # Index dictionary (Key is the value returned by _create_cache_key()
        # (which is based on the header file name) and value is an
        # index_entry_t object)
        self.__index = {}

        # Flag that indicates whether the index was modified
        self.__modified_flag = False

        # Check if dir refers to an existing file...
        if os.path.isfile(self.__dir):
            raise ValueError((
                "Cannot use %s as cache directory. There is already a file " +
                "with that name.") % self.__dir)

        # Load the cache or create the cache directory...
        if os.path.isdir(self.__dir):
            self._load()
        else:
            # Create the cache directory...
            os.mkdir(self.__dir)

    def flush(self):
        """Save the index table to disk."""
        self._save()

    def update(self, source_file, configuration, declarations, included_files):
        """Replace a cache entry by a new value.

        :param source_file: a C++ source file name.
        :type source_file: str
        :param configuration: configuration  object.
        :type configuration: :class:`xml_generator_configuration_t`
        :param declarations: declarations contained in the `source_file`
        :type declarations: pickable object
        :param included_files: included files
        :type included_files: list of str
        """

        # Normlize all paths...
        source_file = os.path.normpath(source_file)
        included_files = [os.path.normpath(p) for p in included_files]

        # Create the list of dependent files. This is the included_files list
        # + the source file. Duplicate names are removed.
        dependent_files = {}
        for name in [source_file] + included_files:
            dependent_files[name] = 1

        key = self._create_cache_key(source_file)
        # Remove an existing entry (if there is one)
        # After calling this method, it is guaranteed that __index[key]
        # does not exist anymore.
        self._remove_entry(source_file, key)

        # Create a new entry...

        # Create the sigs of all dependent files...
        filesigs = []
        for filename in list(dependent_files.keys()):
            id_, sig = self.__filename_rep.acquire_filename(filename)
            filesigs.append((id_, sig))

        configsig = self._create_config_signature(configuration)
        entry = index_entry_t(filesigs, configsig)
        self.__index[key] = entry
        self.__modified_flag = True

        # Write the declarations into the cache file...
        cachefilename = self._create_cache_filename(source_file)
        self._write_file(cachefilename, declarations)

    def cached_value(self, source_file, configuration):
        """Return the cached declarations or None.

        :param source_file: Header file name
        :type source_file: str
        :param configuration: Configuration object
        :type configuration: :class:`parser.xml_generator_configuration_t`
        :rtype: Cached declarations or None
        """

        # Check if the cache contains an entry for source_file
        key = self._create_cache_key(source_file)
        entry = self.__index.get(key)
        if entry is None:
            # print "CACHE: %s: Not cached"%source_file
            return None

        # Check if the entry is still valid. It is not valid if:
        #  - the source_file has been updated
        #  - the configuration object has changed (i.e. the header is parsed
        #    by gccxml with different settings which may influence the
        #    declarations)
        #  - the included files have been updated
        #    (this list is part of the cache entry as it cannot be known
        #    by the caller when cached_value() is called. It was instead
        #    passed to update())

        # Check if the config is different...
        configsig = self._create_config_signature(configuration)
        if configsig != entry.configsig:
            # print "CACHE: %s: Config mismatch"%source_file
            return None

        # Check if any of the dependent files has been modified...
        for id_, sig in entry.filesigs:
            if self.__filename_rep.is_file_modified(id_, sig):
                # print "CACHE: %s: Entry not up to date"%source_file
                return None

        # Load and return the cached declarations
        cachefilename = self._create_cache_filename(source_file)
        decls = self._read_file(cachefilename)

        # print "CACHE: Using cached decls for",source_file
        return decls

    def _load(self):
        """Load the cache.

        Loads the `index.dat` file, which contains the index table and the
        file name repository.

        This method is called by the :meth:`__init__`
        """

        indexfilename = os.path.join(self.__dir, "index.dat")
        if os.path.exists(indexfilename):
            data = self._read_file(indexfilename)
            self.__index = data[0]
            self.__filename_rep = data[1]
            if self.__filename_rep._sha1_sigs != self.__sha1_sigs:
                print((
                    "CACHE: Warning: sha1_sigs stored in the cache is set " +
                    "to %s.") % self.__filename_rep._sha1_sigs)
                print("Please remove the cache to change this setting.")
                self.__sha1_sigs = self.__filename_rep._sha1_sigs
        else:
            self.__index = {}
            self.__filename_rep = filename_repository_t(self.__sha1_sigs)

        self.__modified_flag = False

    def _save(self):
        """
        save the cache index, in case it was modified.

        Saves the index table and the file name repository in the file
        `index.dat`
        """

        if self.__modified_flag:
            self.__filename_rep.update_id_counter()
            indexfilename = os.path.join(self.__dir, "index.dat")
            self._write_file(
                indexfilename,
                (self.__index,
                 self.__filename_rep))

            self.__modified_flag = False

    def _read_file(self, filename):
        """
        read a Python object from a cache file.

        Reads a pickled object from disk and returns it.

        :param filename: Name of the file that should be read.
        :type filename: str
        :rtype: object
        """

        if self.__compression:
            f = gzip.GzipFile(filename, "rb")
        else:
            f = open(filename, "rb")
        res = pickle.load(f)
        f.close()
        return res

    def _write_file(self, filename, data):
        """Write a data item into a file.

        The data object is written to a file using the pickle mechanism.

        :param filename: Output file name
        :type filename: str
        :param data: A Python object that will be pickled
        """

        if self.__compression:
            f = gzip.GzipFile(filename, "wb")
        else:
            f = open(filename, "wb")
        pickle.dump(data, f, pickle.HIGHEST_PROTOCOL)
        f.close()

    def _remove_entry(self, source_file, key):
        """Remove an entry from the cache.

        source_file is the name of the header and key is its corresponding
        cache key (obtained by a call to :meth:_create_cache_key ).
        The entry is removed from the index table, any referenced file
        name is released and the cache file is deleted.

        If key references a non-existing entry, the method returns
        immediately.

        :param source_file: Header file name
        :type source_file: str
        :param key: Key value for the specified header file
        :type key: hash table object
        """

        entry = self.__index.get(key)
        if entry is None:
            return

        # Release the referenced files...
        for id_, _ in entry.filesigs:
            self.__filename_rep.release_filename(id_)

        # Remove the cache entry...
        del self.__index[key]
        self.__modified_flag = True

        # Delete the corresponding cache file...
        cachefilename = self._create_cache_filename(source_file)
        try:
            os.remove(cachefilename)
        except OSError as e:
            print("Could not remove cache file (%s)" % e)

    @staticmethod
    def _create_cache_key(source_file):
        """
        return the cache key for a header file.

        :param source_file: Header file name
        :type source_file: str
        :rtype: str
        """
        path, name = os.path.split(source_file)
        return name + str(hash(path))

    def _create_cache_filename(self, source_file):
        """
        return the cache file name for a header file.

        :param source_file: Header file name
        :type source_file: str
        :rtype: str
        """
        res = self._create_cache_key(source_file) + ".cache"
        return os.path.join(self.__dir, res)

    @staticmethod
    def _create_config_signature(config):
        """
        return the signature for a config object.

        The signature is computed as sha1 digest of the contents of
        working_directory, include_paths, define_symbols and
        undefine_symbols.

        :param config: Configuration object
        :type config: :class:`parser.xml_generator_configuration_t`
        :rtype: str
        """
        m = hashlib.sha1()
        m.update(config.working_directory.encode("utf-8"))
        for p in config.include_paths:
            m.update(p.encode("utf-8"))
        for p in config.define_symbols:
            m.update(p.encode("utf-8"))
        for p in config.undefine_symbols:
            m.update(p.encode("utf-8"))
        for p in config.cflags:
            m.update(p.encode("utf-8"))
        return m.digest()


class filename_entry_t(object):

    """This is a record stored in the filename_repository_t class.

    The class is an internal class used in the implementation of the
    filename_repository_t class and it just serves as a container for
    the file name and the reference count.
    """

    def __init__(self, filename):
        """Constructor.

        The reference count is initially set to 0.
        """
        # Filename
        self.filename = filename
        # Reference count
        self.refcount = 0

        # Cached signature value for the file.
        # If sig_valid flag is False, the signature still has to be computed,
        # otherwise the cached value can be used.
        # These attributes must not be pickled!
        self.sig_valid = False
        self.signature = None

    def __getstate__(self):
        # Only pickle filename and refcount
        return self.filename, self.refcount

    def __setstate__(self, state):
        self.filename, self.refcount = state
        self.sig_valid = False
        self.signature = None

    def inc_ref_count(self):
        """Increase the reference count by 1."""

        self.refcount += 1

    def dec_ref_count(self):
        """Decrease the reference count by 1 and return the new count."""

        self.refcount -= 1
        return self.refcount


class filename_repository_t(object):

    """File name repository.

    This class stores file names and can check whether a file has been
    modified or not since a previous call.
    A file name is stored by calling acquire_filename() which returns
    an ID and a signature of the file. The signature can later be used
    to check if the file was modified by calling is_file_modified().
    If the file name is no longer required release_filename() should be
    called so that the entry can be removed from the repository.
    """

    def __init__(self, sha1_sigs):
        """Constructor.
        """

        # Flag that determines whether the signature is a sha1 digest or
        # the modification time
        # (this flag is passed to the filename_repository_t class)
        self._sha1_sigs = sha1_sigs

        # ID lookup table (key: filename / value: id_)
        self.__id_lut = {}

        # Entry dictionary (key: id_ / value: filename_entry_t)
        # This dictionary contains the actual data.
        # It must always hold that each entry in __entries has a corresponding
        # entry in __id_lut (i.e. the keys in __id_lut must be the names
        # stored in __entries)
        self.__entries = {}

        # A counter for new ids
        self.__next_id = 1

    def acquire_filename(self, name):
        """Acquire a file name and return its id and its signature.
        """

        id_ = self.__id_lut.get(name)
        # Is this a new entry?
        if id_ is None:
            # then create one...
            id_ = self.__next_id
            self.__next_id += 1
            self.__id_lut[name] = id_
            entry = filename_entry_t(name)
            self.__entries[id_] = entry
        else:
            # otherwise obtain the entry...
            entry = self.__entries[id_]

        entry.inc_ref_count()
        return id_, self._get_signature(entry)

    def release_filename(self, id_):
        """Release a file name.
        """

        entry = self.__entries.get(id_)
        if entry is None:
            raise ValueError("Invalid filename id (%d)" % id_)

        # Decrease reference count and check if the entry has to be removed...
        if entry.dec_ref_count() == 0:
            del self.__entries[id_]
            del self.__id_lut[entry.filename]

    def is_file_modified(self, id_, signature):
        """Check if the file referred to by `id_` has been modified.
        """

        entry = self.__entries.get(id_)
        if entry is None:
            raise ValueError("Invalid filename id_ (%d)" % id_)

        # Is the signature already known?
        if entry.sig_valid:
            # use the cached signature
            filesig = entry.signature
        else:
            # compute the signature and store it
            filesig = self._get_signature(entry)
            entry.signature = filesig
            entry.sig_valid = True

        return filesig != signature

    def update_id_counter(self):
        """Update the `id_` counter so that it doesn't grow forever.
        """

        if not self.__entries:
            self.__next_id = 1
        else:
            self.__next_id = max(self.__entries.keys()) + 1

    def _get_signature(self, entry):
        """Return the signature of the file stored in entry.
        """

        if self._sha1_sigs:
            # return sha1 digest of the file content...
            if not os.path.exists(entry.filename):
                return None
            try:
                with open(entry.filename, "r") as f:
                    data = f.read()
                    return hashlib.sha1(data.encode("utf-8")).digest()
            except IOError as e:
                print("Cannot determine sha1 digest:", e)
                return None
        else:
            # return file modification date...
            try:
                return os.path.getmtime(entry.filename)
            except OSError:
                return None

    def _dump(self):  # pragma: no cover
        """
        Dump contents for debugging/testing.
        """

        print(70 * "-")
        print("ID lookup table:")
        for name in self.__id_lut:
            id_ = self.__id_lut[name]
            print("  %s -> %d" % (name, id_))

        print(70 * "-")
        print("%-4s %-60s %s" % ("ID", "Filename", "Refcount"))
        print(70 * "-")
        for id_ in self.__entries:
            entry = self.__entries[id_]
            print("%04d %-60s %d" % (id_, entry.filename, entry.refcount))



# Copyright 2014-2017 Insight Software Consortium.
# Copyright 2004-2009 Roman Yakovenko.
# Distributed under the Boost Software License, Version 1.0.
# See http://www.boost.org/LICENSE_1_0.txt

from . import calldef
from . import declaration_utils
from . import cpptypes
from . import calldef_types


# Second level in hierarchy of calldef
class member_calldef_t(calldef.calldef_t):

    """base class for "callable" declarations that defined within
    C++ class or struct"""

    def __init__(
            self,
            virtuality=None,
            has_const=None,
            has_static=None,
            *args,
            **keywords):
        calldef.calldef_t.__init__(self, *args, **keywords)
        self._virtuality = virtuality
        self._has_const = has_const
        self._has_static = has_static

    def __str__(self):
        # Get the full name of the calldef...
        name = declaration_utils.full_name(self)
        if name[:2] == "::":
            name = name[2:]
        # Add the arguments...
        args = [str(a) for a in self.arguments]
        res = "%s(%s)" % (name, ", ".join(args))
        # Add the return type...
        if self.return_type is not None:
            res = "%s %s" % (self.return_type, res)
        # const?
        if self.has_const:
            res += " const"
        # static?
        if self.has_static:
            res = "static " + res
        # Append the declaration class
        cls = self.__class__.__name__
        if cls[-2:] == "_t":
            cls = cls[:-2]
        cls = cls.replace('_', ' ')
        return "%s [%s]" % (res, cls)

    def _get__cmp__call_items(self):
        """implementation details"""
        return [self.virtuality, self.has_static, self.has_const]

    def __eq__(self, other):
        if not calldef.calldef_t.__eq__(self, other):
            return False
        return self.virtuality == other.virtuality \
            and self.has_static == other.has_static \
            and self.has_const == other.has_const

    __hash__ = calldef.calldef_t.__hash__

    @property
    def virtuality(self):
        """Describes the "virtuality" of the member (as defined by the
            string constants in the class :class:VIRTUALITY_TYPES).
            @type: str"""
        return self._virtuality

    @virtuality.setter
    def virtuality(self, virtuality):
        assert virtuality in calldef_types.VIRTUALITY_TYPES.ALL
        self._virtuality = virtuality

    @property
    def access_type(self):
        """Return the access type of the member (as defined by the
            string constants in the class :class:ACCESS_TYPES.
            @type: str"""
        return self.parent.find_out_member_access_type(self)

    @property
    def has_const(self):
        """describes, whether "callable" has const modifier or not"""
        return self._has_const

    @has_const.setter
    def has_const(self, has_const):
        self._has_const = has_const

    @property
    def has_static(self):
        """describes, whether "callable" has static modifier or not"""
        return self._has_static

    @has_static.setter
    def has_static(self, has_static):
        self._has_static = has_static

    def function_type(self):
        """returns function type. See :class:`type_t` hierarchy"""
        if self.has_static:
            return cpptypes.free_function_type_t(
                return_type=self.return_type,
                arguments_types=[arg.decl_type for arg in self.arguments])

        return cpptypes.member_function_type_t(
            class_inst=self.parent,
            return_type=self.return_type,
            arguments_types=[arg.decl_type for arg in self.arguments],
            has_const=self.has_const)

    def create_decl_string(self, with_defaults=True):
        f_type = self.function_type()
        if with_defaults:
            return f_type.decl_string

        return f_type.partial_decl_string

    def guess_calling_convention(self):
        if self.has_static:
            return calldef_types.CALLING_CONVENTION_TYPES.SYSTEM_DEFAULT

        return calldef_types.CALLING_CONVENTION_TYPES.THISCALL


class operator_t(calldef.calldef_t):
    """
    Base class for "operator" declarations.

    Operators are constructs which behave like functions. Therefore,
    operator_t has calldef_t as parent class.
    """
    OPERATOR_WORD_LEN = len('operator')

    def __init__(self, *args, **keywords):
        calldef.calldef_t.__init__(self, *args, **keywords)

    def _get__cmp__call_items(self):
        raise NotImplementedError()

    @property
    def symbol(self):
        """operator's symbol. For example: operator+, symbol is equal to '+'"""
        return self.name[operator_t.OPERATOR_WORD_LEN:].strip()


# Third level in hierarchy of calldef
class member_function_t(member_calldef_t):

    """describes member function declaration"""

    def __init__(self, *args, **keywords):
        member_calldef_t.__init__(self, *args, **keywords)


class constructor_t(member_calldef_t):

    """describes constructor declaration"""

    def __init__(self, *args, **keywords):
        member_calldef_t.__init__(self, *args, **keywords)
        self._explicit = True

    @property
    def explicit(self):
        """True, if constructor has "explicit" keyword, False otherwise
            @type: bool"""
        return self._explicit

    @explicit.setter
    def explicit(self, explicit):
        self._explicit = explicit in [True, '1']

    def __str__(self):
        # Get the full name of the calldef...
        name = declaration_utils.full_name(self)
        if name[:2] == "::":
            name = name[2:]
        # Add the arguments...
        args = [str(a) for a in self.arguments]
        res = "%s(%s)" % (name, ", ".join(args))
        # Append the declaration class
        cls = 'constructor'
        return "%s [%s]" % (res, cls)


class destructor_t(member_calldef_t):

    """describes deconstructor declaration"""

    def __init__(self, *args, **keywords):
        member_calldef_t.__init__(self, *args, **keywords)


class member_operator_t(member_calldef_t, operator_t):

    """describes member operator declaration"""

    def __init__(self, *args, **keywords):
        member_calldef_t.__init__(self, *args, **keywords)
        operator_t.__init__(self, *args, **keywords)
        self.__class_types = None


class casting_operator_t(member_calldef_t, operator_t):

    """describes casting operator declaration"""

    def __init__(self, *args, **keywords):
        member_calldef_t.__init__(self, *args, **keywords)
        operator_t.__init__(self, *args, **keywords)



# Copyright 2014-2017 Insight Software Consortium.
# Copyright 2004-2009 Roman Yakovenko.
# Distributed under the Boost Software License, Version 1.0.
# See http://www.boost.org/LICENSE_1_0.txt

import os
import stat

from . import autoconfig

from pygccxml import utils
from pygccxml import parser

TEST_FILES = [
    "core_types.hpp",
]


def test_cached_source_file():

    config = autoconfig.cxx_parsers_cfg.config.clone()

    fconfig = parser.file_configuration_t(
        data=TEST_FILES[0],
        content_type=parser.CONTENT_TYPE.CACHED_SOURCE_FILE)
    try:
        prj_reader = parser.project_reader_t(config)
        prj_reader.read_files(
            [fconfig],
            compilation_mode=parser.COMPILATION_MODE.FILE_BY_FILE)
        assert os.path.exists(fconfig.cached_source_file)
        mtime1 = os.stat(fconfig.cached_source_file)[stat.ST_MTIME]
        prj_reader.read_files(
            [fconfig],
            compilation_mode=parser.COMPILATION_MODE.FILE_BY_FILE)
        mtime2 = os.stat(fconfig.cached_source_file)[stat.ST_MTIME]
        assert mtime1 == mtime2
    finally:
        utils.remove_file_no_raise(fconfig.cached_source_file, config)



# Copyright 2014-2017 Insight Software Consortium.
# Copyright 2004-2009 Roman Yakovenko.
# Distributed under the Boost Software License, Version 1.0.
# See http://www.boost.org/LICENSE_1_0.txt

import os

from . import autoconfig

from pygccxml import parser
from pygccxml import declarations


__code = os.linesep.join([
    'struct a{};',
    'struct b{};',
    'struct c{};',
    'struct d : public a{};',
    'struct e : public a, public b{};',
    'struct f{};',
    'struct g : public d, public f{};',
    'struct h : public f{};',
    'struct i : public h, public g{};'])


__recursive_bases = {
    'a': set(),
    'b': set(),
    'c': set(),
    'd': {'a'},
    'e': {'a', 'b'},
    'f': set(),
    'g': {'d', 'f', 'a'},
    'h': {'f'},
    'i': {'h', 'g', 'd', 'f', 'a'}}

__recursive_derived = {
    'a': {'d', 'e', 'g', 'i'},
    'b': {'e'},
    'c': set(),
    'd': {'g', 'i'},
    'e': set(),
    'f': {'g', 'h', 'i'},
    'g': {'i'},
    'h': {'i'},
    'i': set()}


def test_recursive_bases():
    config = autoconfig.cxx_parsers_cfg.config.clone()
    src_reader = parser.source_reader_t(config)
    decls = declarations.make_flatten(src_reader.read_string(__code))
    classes = [
        inst for inst in decls if isinstance(inst, declarations.class_t)]
    for class_ in classes:
        assert class_.name in __recursive_bases
        all_bases = class_.recursive_bases
        control_bases = __recursive_bases[class_.name]
        assert len(control_bases) == len(all_bases)
        all_bases_names = [hi.related_class.name for hi in all_bases]
        assert set(all_bases_names) == control_bases


def test_recursive_derived():
    config = autoconfig.cxx_parsers_cfg.config.clone()
    src_reader = parser.source_reader_t(config)
    decls = declarations.make_flatten(src_reader.read_string(__code))
    classes = [
        inst for inst in decls if isinstance(
            inst,
            declarations.class_t)]
    for class_ in classes:
        assert class_.name in __recursive_derived
        all_derived = class_.recursive_derived
        control_derived = __recursive_derived[class_.name]
        assert len(control_derived) == len(all_derived)
        all_derived_names = [hi.related_class.name for hi in all_derived]
        assert set(all_derived_names) == control_derived



# Copyright 2014-2017 Insight Software Consortium.
# Copyright 2004-2009 Roman Yakovenko.
# Distributed under the Boost Software License, Version 1.0.
# See http://www.boost.org/LICENSE_1_0.txt

from . import autoconfig

from pygccxml import parser
from pygccxml import declarations

TEST_FILES1 = ["bit_fields.hpp"]
TEST_FILES2 = ["unnamed_ns_bug.hpp"]


def test_namespace_matcher_get_single():
    config = autoconfig.cxx_parsers_cfg.config.clone()
    decls = parser.parse(TEST_FILES1, config)
    criteria = declarations.namespace_matcher_t(name='bit_fields')
    declarations.matcher.get_single(criteria, decls)
    assert str(criteria) == '(decl type==namespace_t) and (name==bit_fields)'


def test_namespace_matcher_allow_empty():
    config = autoconfig.cxx_parsers_cfg.config.clone()
    decls = parser.parse(TEST_FILES1, config)
    global_ns = declarations.get_global_namespace(decls)
    assert 0 == len(global_ns.namespaces('does not exist', allow_empty=True))


def test_namespace_matcher_upper():
    config = autoconfig.cxx_parsers_cfg.config.clone()
    decls = parser.parse(TEST_FILES2, config)
    declarations.matcher.get_single(
        declarations.namespace_matcher_t(name='::'), decls)



# Copyright 2014-2017 Insight Software Consortium.
# Copyright 2004-2009 Roman Yakovenko.
# Distributed under the Boost Software License, Version 1.0.
# See http://www.boost.org/LICENSE_1_0.txt

"""
defines few algorithms, that deals with different C++ type properties

Are you aware of `boost::type_traits <http://www.boost.org/doc/libs/1_37_0/
libs/type_traits/doc/html/boost_typetraits/intro.html>`_
library? pygccxml implements the same functionality.

This module contains a set of very specific traits functions\\classes, each of
which encapsulate a single trait from the C++ type system. For example:
* is a type a pointer or a reference type ?
* does a type have a trivial constructor ?
* does a type have a const-qualifier ?

"""

from . import cpptypes
from . import typedef
from .. import utils


def __remove_alias(type_):
    """
    Implementation detail.

    Args:
        type_ (type_t): type

    Returns:
        type_t: the type associated to the inputted type
    """
    if isinstance(type_, cpptypes.declarated_t) and \
            isinstance(type_.declaration, typedef.typedef_t):
        return __remove_alias(type_.declaration.decl_type)
    if isinstance(type_, cpptypes.compound_t):
        type_.base = __remove_alias(type_.base)
        return type_
    return type_


def remove_alias(type_):
    """
    Returns `type_t` without typedef

    Args:
        type_ (type_t | declaration_t): type or declaration

    Returns:
        type_t: the type associated to the inputted declaration
    """
    if isinstance(type_, cpptypes.type_t):
        type_ref = type_
    elif isinstance(type_, typedef.typedef_t):
        type_ref = type_.decl_type
    else:
        # Not a valid input, just return it
        return type_
    if type_ref.cache.remove_alias:
        return type_ref.cache.remove_alias
    no_alias = __remove_alias(type_ref.clone())
    type_ref.cache.remove_alias = no_alias
    return no_alias


def decompose_type(tp):
    """
    Implementation detail
    """
    if isinstance(tp, cpptypes.compound_t):
        return [tp] + decompose_type(tp.base)
    elif isinstance(tp, typedef.typedef_t):
        return decompose_type(tp.decl_type)
    elif isinstance(tp, cpptypes.declarated_t) and \
            isinstance(tp.declaration, typedef.typedef_t):
        return decompose_type(tp.declaration.decl_type)

    return [tp]


def decompose_class(type_):
    """implementation details"""
    types = decompose_type(type_)
    return [tp.__class__ for tp in types]


def base_type(type_):
    """returns base type.

    For `const int` will return `int`
    """
    types = decompose_type(type_)
    return types[-1]


def _create_cv_types(base):
    """
    Implementation detail.

    """
    return (
        [base,
         cpptypes.const_t(base),
         cpptypes.volatile_t(base),
         cpptypes.volatile_t(cpptypes.const_t(base))]
    )


# Some tuples containing combinations of different types
# These are created once the module is loaded, so that when they are used
# they do not need to be re-created.
_void_def = _create_cv_types(cpptypes.void_t())
_bool_def = _create_cv_types(cpptypes.bool_t())
_float_def = (
    _create_cv_types(cpptypes.float_t()) +
    _create_cv_types(cpptypes.double_t()) +
    _create_cv_types(cpptypes.long_double_t()))
_integral_def = (
    _create_cv_types(cpptypes.char_t()) +
    _create_cv_types(cpptypes.unsigned_char_t()) +
    _create_cv_types(cpptypes.signed_char_t()) +
    _create_cv_types(cpptypes.wchar_t()) +
    _create_cv_types(cpptypes.short_int_t()) +
    _create_cv_types(cpptypes.short_unsigned_int_t()) +
    _create_cv_types(cpptypes.bool_t()) +
    _create_cv_types(cpptypes.int_t()) +
    _create_cv_types(cpptypes.unsigned_int_t()) +
    _create_cv_types(cpptypes.long_int_t()) +
    _create_cv_types(cpptypes.long_unsigned_int_t()) +
    _create_cv_types(cpptypes.long_long_int_t()) +
    _create_cv_types(cpptypes.long_long_unsigned_int_t()) +
    _create_cv_types(cpptypes.int128_t()) +
    _create_cv_types(cpptypes.uint128_t()))


def does_match_definition(given, main, secondary):
    """implementation details"""
    assert isinstance(secondary, tuple)
    assert len(secondary) == 2  # general solution could be provided
    types = decompose_type(given)

    if isinstance(types[0], main):
        return True

    if len(types) >= 2:
        cond1 = isinstance(types[0], main)
        cond2 = isinstance(types[1], secondary)
        cond3 = isinstance(types[1], main)
        cond4 = isinstance(types[0], secondary)
        if (cond1 and cond2) or (cond3 and cond4):
            return True

        if len(types) >= 3:
            classes = set([tp.__class__ for tp in types[:3]])
            desired = set([main] + list(secondary))
            diff = classes.symmetric_difference(desired)
            if not diff:
                return True
            if len(diff) == 2:
                items = list(diff)
                return (
                    issubclass(
                        items[0], items[1]) or issubclass(items[1], items[0]))
            return False
    else:
        return False


def is_bool(type_):
    """
    Check if type is of boolean type.

    Args:
        type_ (type_t): The type to be checked

    Returns:
        bool: True if type is a boolean, False otherwise.
    """
    return remove_alias(type_) in _bool_def


def is_void(type_):
    """
    Check if type is of void type.

    Args:
        type_ (type_t): The type to be checked

    Returns:
        bool: True if type is void, False otherwise.
    """
    return remove_alias(type_) in _void_def


def is_void_pointer(type_):
    """returns True, if type represents `void*`, False otherwise"""
    return is_same(type_, cpptypes.pointer_t(cpptypes.void_t()))


def is_integral(type_):
    """
    Check if type is a C++ integral type

    Args:
        type_ (type_t): The type to be checked

    Returns:
        bool: True if type is a C++ integral type, False otherwise.
    """
    return remove_alias(type_) in _integral_def


def is_floating_point(type_):
    """returns True, if type represents C++ floating point type,
    False otherwise"""

    return remove_alias(type_) in _float_def


def is_arithmetic(type_):
    """returns True, if type represents C++ integral or floating point type,
    False otherwise"""
    return is_integral(type_) or is_floating_point(type_)


def is_pointer(type_):
    """returns True, if type represents C++ pointer type, False otherwise"""
    return does_match_definition(type_,
                                 cpptypes.pointer_t,
                                 (cpptypes.const_t, cpptypes.volatile_t)) \
        or does_match_definition(type_,
                                 cpptypes.pointer_t,
                                 (cpptypes.volatile_t, cpptypes.const_t))


def is_calldef_pointer(type_):
    """returns True, if type represents pointer to free/member function,
    False otherwise"""
    if not is_pointer(type_):
        return False
    nake_type = remove_alias(type_)
    nake_type = remove_cv(nake_type)
    return isinstance(nake_type, cpptypes.compound_t) \
        and isinstance(nake_type.base, cpptypes.calldef_type_t)


def remove_pointer(type_):
    """removes pointer from the type definition

    If type is not pointer type, it will be returned as is.
    """
    nake_type = remove_alias(type_)
    if not is_pointer(nake_type):
        return type_
    elif isinstance(nake_type, cpptypes.volatile_t) and \
            isinstance(nake_type.base, cpptypes.pointer_t):
        return cpptypes.volatile_t(nake_type.base.base)
    elif isinstance(nake_type, cpptypes.const_t) and \
            isinstance(nake_type.base, cpptypes.pointer_t):
        return cpptypes.const_t(nake_type.base.base)
    elif isinstance(nake_type, cpptypes.volatile_t) \
            and isinstance(nake_type.base, cpptypes.const_t) \
            and isinstance(nake_type.base.base, cpptypes.pointer_t):
        return (
            cpptypes.volatile_t(cpptypes.const_t(nake_type.base.base.base))
        )

    return nake_type.base


def is_reference(type_):
    """returns True, if type represents C++ reference type, False otherwise"""
    nake_type = remove_alias(type_)
    return isinstance(nake_type, cpptypes.reference_t)


def is_array(type_):
    """returns True, if type represents C++ array type, False otherwise"""
    nake_type = remove_alias(type_)
    nake_type = remove_reference(nake_type)
    nake_type = remove_cv(nake_type)
    return isinstance(nake_type, cpptypes.array_t)


def array_size(type_):
    """returns array size"""
    nake_type = remove_alias(type_)
    nake_type = remove_reference(nake_type)
    nake_type = remove_cv(nake_type)
    assert isinstance(nake_type, cpptypes.array_t)
    return nake_type.size


def array_item_type(type_):
    """returns array item type"""
    if is_array(type_):
        type_ = remove_alias(type_)
        type_ = remove_cv(type_)
        return type_.base
    elif is_pointer(type_):
        return remove_pointer(type_)
    else:
        raise RuntimeError(
            "array_item_type functions takes as argument array or pointer " +
            "types")


def remove_reference(type_):
    """removes reference from the type definition

    If type is not reference type, it will be returned as is.
    """
    nake_type = remove_alias(type_)
    if not is_reference(nake_type):
        return type_

    return nake_type.base


def is_const(type_):
    """returns True, if type represents C++ const type, False otherwise"""
    nake_type = remove_alias(type_)
    if isinstance(nake_type, cpptypes.const_t):
        return True
    elif isinstance(nake_type, cpptypes.volatile_t):
        return is_const(nake_type.base)
    elif isinstance(nake_type, cpptypes.array_t):
        return is_const(nake_type.base)
    return False


def remove_const(type_):
    """removes const from the type definition

    If type is not const type, it will be returned as is
    """

    nake_type = remove_alias(type_)
    if not is_const(nake_type):
        return type_
    else:
        # Handling for const and volatile qualified types. There is a
        # difference in behavior between GCCXML and CastXML for cv-qual arrays.
        # GCCXML produces the following nesting of types:
        #       ->  volatile_t(const_t(array_t))
        # while CastXML produces the following nesting:
        #       ->  array_t(volatile_t(const_t))
        # For both cases, we must unwrap the types, remove const_t, and add
        # back the outer layers
        if isinstance(nake_type, cpptypes.array_t):
            is_v = is_volatile(nake_type)
            if is_v:
                result_type = nake_type.base.base.base
            else:
                result_type = nake_type.base.base
            if is_v:
                result_type = cpptypes.volatile_t(result_type)
            return cpptypes.array_t(result_type, nake_type.size)

        elif isinstance(nake_type, cpptypes.volatile_t):
            return cpptypes.volatile_t(nake_type.base.base)

        return nake_type.base


def remove_declarated(type_):
    """removes type-declaration class-binder :class:`declarated_t` from
    the `type_`

    If `type_` is not :class:`declarated_t`, it will be returned as is
    """
    type_ = remove_alias(type_)
    if isinstance(type_, cpptypes.elaborated_t):
        type_ = type_.base
    if isinstance(type_, cpptypes.declarated_t):
        type_ = type_.declaration
    return type_


def is_same(type1, type2):
    """returns True, if type1 and type2 are same types"""
    nake_type1 = remove_declarated(type1)
    nake_type2 = remove_declarated(type2)
    return nake_type1 == nake_type2


def is_elaborated(type_):
    """returns True, if type represents C++ elaborated type, False otherwise"""
    nake_type = remove_alias(type_)
    if isinstance(nake_type, cpptypes.elaborated_t):
        return True
    elif isinstance(nake_type, cpptypes.reference_t):
        return is_elaborated(nake_type.base)
    elif isinstance(nake_type, cpptypes.pointer_t):
        return is_elaborated(nake_type.base)
    elif isinstance(nake_type, cpptypes.volatile_t):
        return is_elaborated(nake_type.base)
    elif isinstance(nake_type, cpptypes.const_t):
        return is_elaborated(nake_type.base)
    return False


def remove_elaborated(type_):
    """removes type-declaration class-binder :class:`elaborated_t` from
    the `type_`

    If `type_` is not :class:`elaborated_t`, it will be returned as is
    """
    nake_type = remove_alias(type_)
    if not is_elaborated(nake_type):
        return type_
    else:
        if isinstance(type_, cpptypes.elaborated_t):
            type_ = type_.base
    return type_


def is_volatile(type_):
    """returns True, if type represents C++ volatile type, False otherwise"""
    nake_type = remove_alias(type_)
    if isinstance(nake_type, cpptypes.volatile_t):
        return True
    elif isinstance(nake_type, cpptypes.const_t):
        return is_volatile(nake_type.base)
    elif isinstance(nake_type, cpptypes.array_t):
        return is_volatile(nake_type.base)
    return False


def remove_volatile(type_):
    """removes volatile from the type definition

    If type is not volatile type, it will be returned as is
    """
    nake_type = remove_alias(type_)
    if not is_volatile(nake_type):
        return type_
    else:
        if isinstance(nake_type, cpptypes.array_t):
            is_c = is_const(nake_type)
            if is_c:
                base_type_ = nake_type.base.base.base
            else:
                base_type_ = nake_type.base.base
            result_type = base_type_
            if is_c:
                result_type = cpptypes.const_t(result_type)
            return cpptypes.array_t(result_type, nake_type.size)
        return nake_type.base


def remove_cv(type_):
    """removes const and volatile from the type definition"""

    nake_type = remove_alias(type_)
    if not is_const(nake_type) and not is_volatile(nake_type):
        return type_
    result = nake_type
    if is_const(result):
        result = remove_const(result)
    if is_volatile(result):
        result = remove_volatile(result)
    if is_const(result):
        result = remove_const(result)
    return result


def is_fundamental(type_):
    """returns True, if type represents C++ fundamental type"""
    return does_match_definition(
        type_,
        cpptypes.fundamental_t,
        (cpptypes.const_t, cpptypes.volatile_t)) \
        or does_match_definition(
            type_,
            cpptypes.fundamental_t,
            (cpptypes.volatile_t, cpptypes.const_t))


def _normalize(string):
    return string.replace(' ', '').replace("::std", "std")


def _normalize_equivalences(equivalences):
    return [_normalize(eq) for eq in equivalences]


string_equivalences = [
    (
        'std::basic_string<char, std::char_traits<char>, '
        'std::allocator<char>>'
    ),
    'std::basic_string<char>',
    'std::string'
    ]

wstring_equivalences = [
    (
        'std::basic_string<wchar_t, std::char_traits<wchar_t>, '
        'std::allocator<wchar_t>>'
    ),
    'std::basic_string<wchar_t>',
    'std::wstring'
    ]

ostream_equivalences = [
    'std::basic_ostream<char, std::char_traits<char>>',
    'std::basic_ostream<char>',
    'std::ostream'
    ]

wostream_equivalences = [
    'std::basic_ostream<wchar_t, std::char_traits<wchar_t>>',
    'std::basic_ostream<wchar_t>',
    'std::wostream'
    ]


normalized_string_equivalences = _normalize_equivalences(
    string_equivalences
    )
normalized_wstring_equivalences = _normalize_equivalences(
    wstring_equivalences
    )
normalized_ostream_equivalences = _normalize_equivalences(
    ostream_equivalences
    )
normalized_wostream_equivalences = _normalize_equivalences(
    wostream_equivalences
    )


def is_std_string(type_):
    """
    Returns True, if type represents C++ `std::string`, False otherwise.

    """

    if isinstance(type_, str):
        return _normalize(type_) in normalized_string_equivalences

    type_ = remove_alias(type_)
    type_ = remove_reference(type_)
    type_ = remove_cv(type_)
    return _normalize(type_.decl_string) in normalized_string_equivalences


def is_std_wstring(type_):
    """
    Returns True, if type represents C++ `std::wstring`, False otherwise.

    """

    if isinstance(type_, str):
        return _normalize(type_) in normalized_wstring_equivalences

    type_ = remove_alias(type_)
    type_ = remove_reference(type_)
    type_ = remove_cv(type_)
    return _normalize(type_.decl_string) in normalized_wstring_equivalences


def is_std_ostream(type_):
    """
    Returns True, if type represents C++ std::ostream, False otherwise.

    """

    if isinstance(type_, str):
        return _normalize(type_) in normalized_ostream_equivalences

    type_ = remove_alias(type_)
    type_ = remove_reference(type_)
    type_ = remove_cv(type_)
    return _normalize(type_.decl_string) in normalized_ostream_equivalences


def is_std_wostream(type_):
    """
    Returns True, if type represents C++ std::wostream, False otherwise.

    """

    if isinstance(type_, str):
        return _normalize(type_) in normalized_wostream_equivalences

    type_ = remove_alias(type_)
    type_ = remove_reference(type_)
    type_ = remove_cv(type_)
    return _normalize(type_.decl_string) in normalized_wostream_equivalences



# Copyright 2014-2017 Insight Software Consortium.
# Copyright 2004-2009 Roman Yakovenko.
# Distributed under the Boost Software License, Version 1.0.
# See http://www.boost.org/LICENSE_1_0.txt

import pytest

from . import autoconfig

from pygccxml import parser
from pygccxml import declarations


TEST_FILES = [
    "const_volatile_arg.hpp",
]


@pytest.fixture
def global_ns():
    COMPILATION_MODE = parser.COMPILATION_MODE.ALL_AT_ONCE
    INIT_OPTIMIZER = True
    config = autoconfig.cxx_parsers_cfg.config.clone()
    config.castxml_epic_version = 1
    decls = parser.parse(TEST_FILES, config, COMPILATION_MODE)
    global_ns = declarations.get_global_namespace(decls)
    if INIT_OPTIMIZER:
        global_ns.init_optimizer()
    return global_ns


def test_const_volatile_arg(global_ns):
    f = global_ns.free_function('pygccxml_bug')
    t = f.arguments[0].decl_type
    assert isinstance(t, declarations.pointer_t)
    assert isinstance(t.base, declarations.volatile_t)
    assert isinstance(t.base.base, declarations.const_t)
    assert declarations.is_integral(t.base.base.base)



# Copyright 2014-2017 Insight Software Consortium.
# Copyright 2004-2009 Roman Yakovenko.
# Distributed under the Boost Software License, Version 1.0.
# See http://www.boost.org/LICENSE_1_0.txt

import re


class VIRTUALITY_TYPES(object):

    """class that defines "virtuality" constants"""
    NOT_VIRTUAL = 'not virtual'
    VIRTUAL = 'virtual'
    PURE_VIRTUAL = 'pure virtual'
    ALL = [NOT_VIRTUAL, VIRTUAL, PURE_VIRTUAL]


# preserving backward compatebility
FUNCTION_VIRTUALITY_TYPES = VIRTUALITY_TYPES


class CALLING_CONVENTION_TYPES(object):

    """class that defines "calling convention" constants"""
    UNKNOWN = ''
    CDECL = 'cdecl'
    STDCALL = 'stdcall'
    THISCALL = 'thiscall'
    FASTCALL = 'fastcall'
    SYSTEM_DEFAULT = '<<<system default>>>'

    all = (UNKNOWN, CDECL, STDCALL, THISCALL, FASTCALL, SYSTEM_DEFAULT)

    pattern = re.compile(
        r'.*(?:^|\s)(?:__)?(?P<cc>cdecl|stdcall|thiscall|fastcall)(?:__)?.*')

    @staticmethod
    def extract(text, default=UNKNOWN):
        """extracts calling convention from the text. If the calling convention
        could not be found, the "default"is used"""
        if not text:
            return default
        found = CALLING_CONVENTION_TYPES.pattern.match(text)
        if found:
            return found.group('cc')

        return default



# Copyright 2014-2017 Insight Software Consortium.
# Copyright 2004-2009 Roman Yakovenko.
# Distributed under the Boost Software License, Version 1.0.
# See http://www.boost.org/LICENSE_1_0.txt

"""
template instantiation parser

This module provides functionality necessary to

* :func:`parse <pygccxml.declarations.templates.parse>`
* :func:`split <pygccxml.declarations.templates.split>`
* :func:`join <pygccxml.declarations.templates.join>`
* :func:`normalize <pygccxml.declarations.templates.normalize>`

C++ template instantiations
"""

from . import pattern_parser
from . import declaration_utils

__THE_PARSER = pattern_parser.parser_t('<', '>', ',')


def is_instantiation(decl_string):
    """
    returns True if `decl_string` is template instantiation and False otherwise

    :param decl_string: string that should be checked for pattern presence
    :type decl_string: str

    :rtype: bool
    """
    return __THE_PARSER.has_pattern(decl_string)


def name(decl_string):
    """
    returns name of instantiated template

    :type decl_string: str
    :rtype: str
    """
    return __THE_PARSER.name(decl_string)


def args(decl_string):
    """
    returns list of template arguments

    :type decl_string: `str`
    :rtype: [`str`]
    """
    return __THE_PARSER.args(decl_string)


def split(decl_string):
    """returns (name, [arguments] )"""
    return __THE_PARSER.split(decl_string)


def split_recursive(decl_string):
    """returns [(name, [arguments])]"""
    return __THE_PARSER.split_recursive(decl_string)


def join(name_, args_):
    """returns name< argument_1, argument_2, ..., argument_n >"""
    return __THE_PARSER.join(name_, args_)


def normalize(decl_string):
    """returns `decl_string`, which contains "normalized" spaces

    this functionality allows to implement comparison of 2 different string
    which are actually same: x::y< z > and x::y<z>
    """
    return __THE_PARSER.normalize(decl_string)


def normalize_name(decl):
    """
    Cached variant of normalize

    Args:
        decl (declaration.declaration_t): the declaration

    Returns:
        str: normalized name
    """
    if decl.cache.normalized_name is None:
        decl.cache.normalized_name = normalize(decl.name)
    return decl.cache.normalized_name


def normalize_partial_name(decl):
    """
    Cached variant of normalize

    Args:
        decl (declaration.declaration_t): the declaration

    Returns:
        str: normalized name
    """
    if decl.cache.normalized_partial_name is None:
        decl.cache.normalized_partial_name = normalize(decl.partial_name)
    return decl.cache.normalized_partial_name


def normalize_full_name_true(decl):
    """
    Cached variant of normalize

    Args:
        decl (declaration.declaration_t): the declaration

    Returns:
        str: normalized name
    """
    if decl.cache.normalized_full_name_true is None:
        decl.cache.normalized_full_name_true = normalize(
            declaration_utils.full_name(decl, with_defaults=True))
    return decl.cache.normalized_full_name_true


def normalize_full_name_false(decl):
    """
    Cached variant of normalize

    Args:
        decl (declaration.declaration_t): the declaration

    Returns:
        str: normalized name
    """
    if decl.cache.normalized_full_name_false is None:
        decl.cache.normalized_full_name_false = normalize(
            declaration_utils.full_name(decl, with_defaults=False))
    return decl.cache.normalized_full_name_false



# Copyright 2014-2017 Insight Software Consortium.
# Copyright 2004-2009 Roman Yakovenko.
# Distributed under the Boost Software License, Version 1.0.
# See http://www.boost.org/LICENSE_1_0.txt

import pytest

from . import autoconfig

from pygccxml import parser
from pygccxml import declarations

TEST_FILES = ['test_argument_without_name.hpp']


@pytest.fixture
def global_ns():
    COMPILATION_MODE = parser.COMPILATION_MODE.ALL_AT_ONCE
    config = autoconfig.cxx_parsers_cfg.config.clone()
    config.cflags = "-std=c++11"
    decls = parser.parse(TEST_FILES, config, COMPILATION_MODE)
    global_ns = declarations.get_global_namespace(decls)
    global_ns.init_optimizer()
    return global_ns


def test_argument_without_name(global_ns):
    """
    Test passing an object without name to a templated function.

    The test was failing when building the declaration string.
    The declaration string will be 'void (*)(  & )'. If the passed
    object had a name the result would then be 'void (*)(Name & )'.

    See bug #55

    """

    criteria = declarations.calldef_matcher(name="function")
    free_funcs = declarations.matcher.find(criteria, global_ns)
    for free_func in free_funcs:
        decl_string = free_func.create_decl_string(with_defaults=False)
        assert decl_string == "void (*)(  & )"



# Copyright 2014-2017 Insight Software Consortium.
# Copyright 2004-2009 Roman Yakovenko.
# Distributed under the Boost Software License, Version 1.0.
# See http://www.boost.org/LICENSE_1_0.txt

import os
import timeit
import hashlib
try:
    import cPickle as pickle
except ImportError:
    import pickle
from pygccxml import utils
from . import config as cxx_parsers_cfg


def file_signature(filename):
    """
    Return a signature for a file.

    """

    if not os.path.isfile(filename):
        return None
    if not os.path.exists(filename):
        return None

    # Duplicate auto-generated files can be recognized with the sha1 hash.
    sig = hashlib.sha1()
    with open(filename, "rb") as f:
        buf = f.read()
        sig.update(buf)

    return sig.hexdigest()


def configuration_signature(config):
    """
    Return a signature for a configuration (xml_generator_configuration_t)
    object.

    This can then be used as a key in the cache.
    This method must take into account anything about
    a configuration that could cause the declarations generated
    to be different between runs.

    """

    sig = hashlib.sha1()
    if isinstance(config, cxx_parsers_cfg.xml_generator_configuration_t):
        sig.update(str(config.xml_generator_path).encode())
    sig.update(str(config.working_directory).encode('utf-8'))
    if isinstance(config, cxx_parsers_cfg.xml_generator_configuration_t):
        sig.update(str(config.cflags).encode('utf-8'))
    for p in config.include_paths:
        sig.update(str(p).encode('utf-8'))
    for s in config.define_symbols:
        sig.update(str(s).encode('utf-8'))
    for u in config.undefine_symbols:
        sig.update(str(u).encode('utf-8'))
    return sig.hexdigest()


class cache_base_t(object):
    logger = utils.loggers.declarations_cache

    def __init__(self):
        object.__init__(self)

    def flush(self):
        """ Flush (write out) the cache to disk if needed. """

        raise NotImplementedError()

    def update(self, source_file, configuration, declarations, included_files):
        """
        update cache entry

        :param source_file: path to the C++ source file being parsed
        :param configuration: configuration used in
               parsing :class:`xml_generator_configuration_t`
        :param declarations: declaration tree found when parsing
        :param included_files: files included by parsing.
        """

        raise NotImplementedError()

    def cached_value(self, source_file, configuration):
        """
        Return declarations, we have cached, for the source_file and the
        given configuration.

        :param source_file: path to the C++ source file being parsed.
        :param configuration: configuration that was used for parsing.

        """

        raise NotImplementedError()


class record_t(object):

    def __init__(
            self,
            source_signature,
            config_signature,
            included_files,
            included_files_signature,
            declarations):
        self.__source_signature = source_signature
        self.__config_signature = config_signature
        self.__included_files = included_files
        self.__included_files_signature = included_files_signature
        self.__declarations = declarations
        self.__was_hit = True  # Track if there was a cache hit

    @property
    def was_hit(self):
        return self.__was_hit

    @was_hit.setter
    def was_hit(self, was_hit):
        self.__was_hit = was_hit

    def key(self):
        return self.__source_signature, self.__config_signature

    @staticmethod
    def create_key(source_file, configuration):
        return (
            file_signature(source_file),
            configuration_signature(configuration))

    @property
    def source_signature(self):
        return self.__source_signature

    @property
    def config_signature(self):
        return self.__config_signature

    @property
    def included_files(self):
        return self.__included_files

    @property
    def included_files_signature(self):
        return self.__included_files_signature

    @property
    def declarations(self):
        return self.__declarations


class file_cache_t(cache_base_t):

    """ Cache implementation to store data in a pickled form in a file.
        This class contains some cache logic that keeps track of which entries
        have been 'hit' in the cache and if an entry has not been hit then
        it is deleted at the time of the flush().  This keeps the cache from
        growing larger when files change and are not used again.
    """

    def __init__(self, name):
        """
        :param name: name of the cache file.
        """

        cache_base_t.__init__(self)
        self.__name = name  # Name of cache file
        # Map record_key to record_t
        self.__cache = self.__load(self.__name)
        self.__needs_flushed = not bool(
            self.__cache)  # If empty then we need to flush
        for entry in self.__cache.values():  # Clear hit flags
            entry.was_hit = False

    @staticmethod
    def __load(file_name):
        """ Load pickled cache from file and return the object. """

        if os.path.exists(file_name) and not os.path.isfile(file_name):
            raise RuntimeError(
                'Cache should be initialized with valid full file name')
        if not os.path.exists(file_name):
            open(file_name, 'w+b').close()
            return {}
        cache_file_obj = open(file_name, 'rb')
        try:
            file_cache_t.logger.info('Loading cache file "%s".', file_name)
            start_time = timeit.default_timer()
            cache = pickle.load(cache_file_obj)
            file_cache_t.logger.debug(
                "Cache file has been loaded in %.1f secs",
                (timeit.default_timer() - start_time))
            file_cache_t.logger.debug(
                "Found cache in file: [%s]  entries: %s",
                file_name, len(list(cache.keys())))
        except (pickle.UnpicklingError, AttributeError, EOFError,
                ImportError, IndexError) as error:
            file_cache_t.logger.exception(
                "Error occurred while reading cache file: %s",
                error)
            cache_file_obj.close()
            file_cache_t.logger.info(
                "Invalid cache file: [%s]  Regenerating.",
                file_name)
            open(file_name, 'w+b').close()   # Create empty file
            cache = {}                       # Empty cache
        finally:
            cache_file_obj.close()
        return cache

    def flush(self):
        # If not marked as needing flushed, then return immediately
        if not self.__needs_flushed:
            self.logger.debug("Cache did not change, ignoring flush.")
            return

        # Remove entries that did not get a cache hit
        num_removed = 0
        for key in list(self.__cache.keys()):
            if not self.__cache[key].was_hit:
                num_removed += 1
                del self.__cache[key]
        if num_removed > 0:
            self.logger.debug(
                "There are %s removed entries from cache.",
                num_removed)
        # Save out the cache to disk
        with open(self.__name, "w+b") as cache_file:
            pickle.dump(self.__cache, cache_file, pickle.HIGHEST_PROTOCOL)

    def update(self, source_file, configuration, declarations, included_files):
        """ Update a cached record with the current key and value contents. """

        record = record_t(
            source_signature=file_signature(source_file),
            config_signature=configuration_signature(configuration),
            included_files=included_files,
            included_files_signature=list(
                map(
                    file_signature,
                    included_files)),
            declarations=declarations)
        # Switched over to holding full record in cache so we don't have
        # to keep creating records in the next method.
        self.__cache[record.key()] = record
        self.__cache[record.key()].was_hit = True
        self.__needs_flushed = True

    def cached_value(self, source_file, configuration):
        """
        Attempt to lookup the cached declarations for the given file and
        configuration.

        Returns None if declaration not found or signature check fails.

        """

        key = record_t.create_key(source_file, configuration)
        if key not in self.__cache:
            return None
        record = self.__cache[key]
        if self.__is_valid_signature(record):
            record.was_hit = True  # Record cache hit
            return record.declarations

        # some file has been changed
        del self.__cache[key]
        return None

    @staticmethod
    def __is_valid_signature(record):
        for index, included_file in enumerate(record.included_files):
            if file_signature(included_file) != \
                    record.included_files_signature[index]:
                return False
        return True


class dummy_cache_t(cache_base_t):

    """
    This is an empty cache object.

    By default no caching is enabled in pygccxml.

    """

    def __init__(self):
        cache_base_t.__init__(self)

    def flush(self):
        pass

    def update(self, source_file, configuration, declarations, included_files):
        pass

    def cached_value(self, source_file, configuration):
        return None



# Copyright 2014-2017 Insight Software Consortium.
# Copyright 2004-2009 Roman Yakovenko.
# Distributed under the Boost Software License, Version 1.0.
# See http://www.boost.org/LICENSE_1_0.txt

import pytest

from . import autoconfig

from pygccxml import parser
from pygccxml import declarations

TEST_FILES = ['declarations_calldef.hpp']


@pytest.fixture
def global_ns():
    COMPILATION_MODE = parser.COMPILATION_MODE.ALL_AT_ONCE
    config = autoconfig.cxx_parsers_cfg.config.clone()
    decls = parser.parse(TEST_FILES, config, COMPILATION_MODE)
    global_ns = declarations.get_global_namespace(decls)
    global_ns.init_optimizer()
    return global_ns


def test_regex(global_ns):
    criteria = declarations.regex_matcher_t(
        'oper.*',
        lambda decl: decl.name)
    operators = declarations.matcher.find(criteria, global_ns)
    operators = [d for d in operators if not d.is_artificial]
    assert len(operators) == 6


def test_access_type(global_ns):
    criteria = declarations.access_type_matcher_t(
        declarations.ACCESS_TYPES.PUBLIC)
    public_members = declarations.matcher.find(criteria, global_ns)
    public_members = [d for d in public_members if not d.is_artificial]

    nbr = len(public_members)
    assert nbr in [17, 21]
    if nbr == 21:
        # We are using llvm 3.9, see bug #32. Make sure the 4 names
        # are still there
        names = ["isa", "flags", "str", "length"]
        for name in names:
            assert names in [mbr.name for mbr in public_members]


def test_or_matcher(global_ns):
    criteria1 = declarations.regex_matcher_t(
        "oper.*",
        lambda decl: decl.name)
    criteria2 = declarations.access_type_matcher_t(
        declarations.ACCESS_TYPES.PUBLIC)
    found = declarations.matcher.find(
        criteria1 | criteria2,
        global_ns)
    found = [d for d in found if not d.is_artificial]
    assert len(found) != 35


def test_and_matcher(global_ns):
    criteria1 = declarations.regex_matcher_t(
        'oper.*',
        lambda decl: decl.name)
    criteria2 = declarations.access_type_matcher_t(
        declarations.ACCESS_TYPES.PUBLIC)
    found = declarations.matcher.find(
        criteria1 & criteria2,
        global_ns)
    found = [d for d in found if not d.is_artificial]
    assert len(found) <= 6


def test_not_matcher(global_ns):
    criteria1 = declarations.regex_matcher_t(
        'oper.*',
        lambda decl: decl.name)
    found = declarations.matcher.find(~(~criteria1), global_ns)
    found = [d for d in found if not d.is_artificial]
    assert len(found) == 6



# Copyright 2014-2017 Insight Software Consortium.
# Copyright 2004-2009 Roman Yakovenko.
# Distributed under the Boost Software License, Version 1.0.
# See http://www.boost.org/LICENSE_1_0.txt

from pygccxml import utils
from pygccxml import declarations
from pygccxml import parser

import os
import sys
import warnings
warnings.simplefilter("error", Warning)
# Find out the file location within the sources tree
this_module_dir_path = os.path.abspath(
    os.path.dirname(sys.modules[__name__].__file__))

# Find out the c++ parser
generator_path, generator_name = utils.find_xml_generator()

# Configure the xml generator
xml_generator_config = parser.xml_generator_configuration_t(
    xml_generator_path=generator_path,
    xml_generator=generator_name)

# The c++ file we want to parse
filename = "example.hpp"
filename = this_module_dir_path + "/" + filename

decls = parser.parse([filename], xml_generator_config)
global_namespace = declarations.get_global_namespace(decls)

function_ptr = global_namespace.variables()[0]

# Print the name of the function pointer
print(function_ptr.name)
# > myFuncPointer

# Print the type of the declaration
print(function_ptr.decl_type)
# > void (*)( int,double )

# Print the real type of the declaration (it's just a pointer)
print(type(function_ptr.decl_type))
# > <class 'pygccxml.declarations.cpptypes.pointer_t'>

# Check if this is a function pointer
print(declarations.is_calldef_pointer(function_ptr.decl_type))
# > True

# Remove the pointer part, to access the function's type
f_type = declarations.remove_pointer(function_ptr.decl_type)

# Print the type
print(type(f_type))
# > <class 'pygccxml.declarations.cpptypes.free_function_type_t'>

# Print the return type and the arguments of the function
print(f_type.return_type)
# > void

# Print the return type and the arguments
print(str(f_type.arguments_types[0]), str(f_type.arguments_types[1]))
# > int, double



# Copyright 2014-2017 Insight Software Consortium.
# Copyright 2004-2009 Roman Yakovenko.
# Distributed under the Boost Software License, Version 1.0.
# See http://www.boost.org/LICENSE_1_0.txt

import pytest

from . import autoconfig

from pygccxml import parser
from pygccxml import declarations

TEST_FILES = [
    "test_copy_constructor.hpp",
]


@pytest.fixture
def global_ns():
    COMPILATION_MODE = parser.COMPILATION_MODE.ALL_AT_ONCE
    INIT_OPTIMIZER = True
    config = autoconfig.cxx_parsers_cfg.config.clone()
    config.castxml_epic_version = 1
    decls = parser.parse(TEST_FILES, config, COMPILATION_MODE)
    global_ns = declarations.get_global_namespace(decls)
    if INIT_OPTIMIZER:
        global_ns.init_optimizer()
    return global_ns


def test_copy_constructor(global_ns):
    """
    Check the is_copy_constructor method.

    This fails when using CastXML, see issue #27.

    """

    tclass = global_ns.class_("test")
    ctors = []
    for decl in tclass.declarations:
        if isinstance(decl, declarations.constructor_t):
            ctors.append(decl)

    # test::test(test const & t0) [copy constructor]
    assert declarations.is_copy_constructor(ctors[0])
    # test::test(float const & t0) [constructor]
    assert not declarations.is_copy_constructor(ctors[1])
    # test::test(myvar t0) [constructor]
    assert not declarations.is_copy_constructor(ctors[2])

    t2class = global_ns.class_("test2")
    ctors = []
    for decl in t2class.declarations:
        if isinstance(decl, declarations.constructor_t):
            ctors.append(decl)

    # test2::test2() [constructor]
    assert not declarations.is_copy_constructor(ctors[0])
    # test2::test2(test2 const & arg0) [copy constructor]
    assert declarations.is_copy_constructor(ctors[1])



# Copyright 2014-2017 Insight Software Consortium.
# Copyright 2004-2009 Roman Yakovenko.
# Distributed under the Boost Software License, Version 1.0.
# See http://www.boost.org/LICENSE_1_0.txt

from . import autoconfig

from pygccxml import parser
from pygccxml import declarations


def test_array_bug_1():
    config = autoconfig.cxx_parsers_cfg.config.clone()
    code = 'int aaaa[2][3][4][5];'
    src_reader = parser.source_reader_t(config)
    global_ns = declarations.get_global_namespace(
        src_reader.read_string(code))
    aaaa_type = global_ns.variable('aaaa').decl_type
    assert 'int [2][3][4][5]' == aaaa_type.decl_string


def test_array_bug_2():
    config = autoconfig.cxx_parsers_cfg.config.clone()
    code = 'int* aaaa[2][3][4][5];'
    src_reader = parser.source_reader_t(config)
    global_ns = declarations.get_global_namespace(
        src_reader.read_string(code))
    aaaa_type = global_ns.variable('aaaa').decl_type
    assert 'int * [2][3][4][5]' == aaaa_type.decl_string


def test_array_bug_3():
    config = autoconfig.cxx_parsers_cfg.config.clone()
    code = 'int aaaa[2];'
    src_reader = parser.source_reader_t(config)
    global_ns = declarations.get_global_namespace(
        src_reader.read_string(code))
    aaaa_type = global_ns.variable('aaaa').decl_type
    assert 'int [2]' == aaaa_type.decl_string


def test_array_bug_4():
    config = autoconfig.cxx_parsers_cfg.config.clone()
    code = 'struct xyz{}; xyz aaaa[2][3];'
    src_reader = parser.source_reader_t(config)
    global_ns = declarations.get_global_namespace(
        src_reader.read_string(code))
    aaaa_type = global_ns.variable('aaaa').decl_type
    assert '::xyz [2][3]' == aaaa_type.decl_string


def test_array_bug_5():
    config = autoconfig.cxx_parsers_cfg.config.clone()
    code = 'char const arr[4] = {};'
    src_reader = parser.source_reader_t(config)
    global_ns = declarations.get_global_namespace(
        src_reader.read_string(code))
    arr_type = global_ns.variable('arr').decl_type
    assert 'char const [4]' == arr_type.decl_string
    assert declarations.is_array(arr_type) is True
    assert declarations.is_const(arr_type) is True


def test_array_bug_6():
    config = autoconfig.cxx_parsers_cfg.config.clone()
    code = 'char volatile arr[4] = {};'
    src_reader = parser.source_reader_t(config)
    global_ns = declarations.get_global_namespace(
        src_reader.read_string(code))
    arr_type = global_ns.variable('arr').decl_type
    assert 'char volatile [4]' == arr_type.decl_string
    assert declarations.is_array(arr_type) is True
    assert declarations.is_volatile(arr_type) is True


def test_array_bug_7():
    config = autoconfig.cxx_parsers_cfg.config.clone()
    code = 'char const volatile arr[4] = {};'
    src_reader = parser.source_reader_t(config)
    global_ns = declarations.get_global_namespace(
        src_reader.read_string(code))
    arr_type = global_ns.variable('arr').decl_type
    assert 'char const volatile [4]' == arr_type.decl_string
    assert declarations.is_array(arr_type) is True
    assert declarations.is_const(arr_type) is True
    assert declarations.is_volatile(arr_type) is True



# Copyright 2014-2017 Insight Software Consortium.
# Copyright 2004-2009 Roman Yakovenko.
# Distributed under the Boost Software License, Version 1.0.
# See http://www.boost.org/LICENSE_1_0.txt

from . import autoconfig

from pygccxml import parser
from pygccxml import declarations


def test_text_reader():
    config = autoconfig.cxx_parsers_cfg.config.clone()

    fconfig = parser.file_configuration_t(
        data='int i;',
        start_with_declarations=None,
        content_type=parser.file_configuration_t.CONTENT_TYPE.TEXT)

    prj_reader = parser.project_reader_t(config)
    decls = prj_reader.read_files(
        [fconfig],
        compilation_mode=parser.COMPILATION_MODE.FILE_BY_FILE)

    var_i = declarations.find_declaration(
        decls, decl_type=declarations.variable_t, name='i')
    assert var_i is not None



# Copyright 2014-2017 Insight Software Consortium.
# Copyright 2004-2009 Roman Yakovenko.
# Distributed under the Boost Software License, Version 1.0.
# See http://www.boost.org/LICENSE_1_0.txt

"""
defines few algorithms, that deals with different properties of functions
"""

from . import calldef_types
from . import calldef_members
from . import type_traits
from . import type_traits_classes


def is_same_return_type(f1, f2):
    # covariant returns
    # The return type of an overriding function shall be either identical to
    # the return type of the overridden function or covariant with the classes
    # of the functions. If a function D::f overrides a function B::f, the
    # return types of the functions are covariant if they satisfy the following
    # criteria:

    # both are pointers to classes or references to classes
    # the class in the return type of B::f is the same class as the class in
    # the return type of D::f or, is an unambiguous direct or indirect base
    # class of the class in the return type of D::f and is accessible in D
    # both pointers or references have the same cv-qualification and the class
    # type in the return type of D::f has the same cv-qualification as or less
    # cv-qualification than the class type in the return type of B::f.

    if f1.__class__ is not f2.__class__:
        # it should be assert
        return False  # 2 different calldef types
    if not isinstance(f1, calldef_members.member_calldef_t):
        # for free functions we compare return types as usual
        return type_traits.is_same(f1.return_type, f2.return_type)
    if f1.virtuality == calldef_types.VIRTUALITY_TYPES.NOT_VIRTUAL \
       or f2.virtuality == calldef_types.VIRTUALITY_TYPES.NOT_VIRTUAL:
        # for non-virtual member functions we compare types as usual
        return type_traits.is_same(f1.return_type, f2.return_type)
    rt1 = f1.return_type
    rt2 = f2.return_type
    if type_traits.is_pointer(rt1) and type_traits.is_pointer(rt2):
        rt1 = type_traits.remove_pointer(rt1)
        rt2 = type_traits.remove_pointer(rt2)
    elif type_traits.is_reference(rt1) and type_traits.is_reference(rt2):
        rt1 = type_traits.remove_reference(rt1)
        rt2 = type_traits.remove_reference(rt2)
    else:
        return type_traits.is_same(f1.return_type, f2.return_type)
    if (type_traits.is_const(rt1) and type_traits.is_const(rt2)) \
       or (not type_traits.is_const(rt1) and not type_traits.is_const(rt2)):
        rt1 = type_traits.remove_const(rt1)
        rt2 = type_traits.remove_const(rt2)
    else:
        return False
    if not type_traits_classes.is_class(rt1) or not \
            type_traits_classes.is_class(rt2):
        return type_traits.is_same(rt1, rt2)

    if type_traits_classes.is_union(rt1) or \
            type_traits_classes.is_union(rt2):
        return type_traits.is_same(rt1, rt2)

    c1 = type_traits_classes.class_traits.get_declaration(rt1)
    c2 = type_traits_classes.class_traits.get_declaration(rt2)
    return type_traits.is_same(c1, c2) \
        or type_traits_classes.is_base_and_derived(c1, c2) \
        or type_traits_classes.is_base_and_derived(c2, c1)


def is_same_function(f1, f2):
    """returns true if f1 and f2 is same function

    Use case: sometimes when user defines some virtual function in base class,
    it overrides it in a derived one. Sometimes we need to know whether two
    member functions is actually same function.
    """
    if f1 is f2:
        return True
    if f1.__class__ is not f2.__class__:
        return False
    if isinstance(f1, calldef_members.member_calldef_t) and \
            f1.has_const != f2.has_const:
        return False
    if f1.name != f2.name:
        return False
    if not is_same_return_type(f1, f2):
        return False
    if len(f1.arguments) != len(f2.arguments):
        return False
    for f1_arg, f2_arg in zip(f1.arguments, f2.arguments):
        if not type_traits.is_same(f1_arg.decl_type, f2_arg.decl_type):
            return False
    return True



from . import type_traits
from . import class_declaration
from . import matchers
from . import cpptypes


def has_public_binary_operator(type_, operator_symbol):
    """returns True, if `type_` has public binary operator, otherwise False"""
    type_ = type_traits.remove_alias(type_)
    type_ = type_traits.remove_cv(type_)
    type_ = type_traits.remove_declarated(type_)
    assert isinstance(type_, class_declaration.class_t)

    if type_traits.is_std_string(type_) or type_traits.is_std_wstring(type_):
        # In some case compare operators of std::basic_string are not
        # instantiated
        return True

    operators = type_.member_operators(
        function=matchers.custom_matcher_t(
            lambda decl: not decl.is_artificial) &
        matchers.access_type_matcher_t('public'),
        symbol=operator_symbol, allow_empty=True, recursive=False)
    if operators:
        return True

    declarated = cpptypes.declarated_t(type_)
    const = cpptypes.const_t(declarated)
    reference = cpptypes.reference_t(const)
    operators = type_.top_parent.operators(
        function=lambda decl: not decl.is_artificial,
        arg_types=[reference, None],
        symbol=operator_symbol,
        allow_empty=True,
        recursive=True)
    if operators:
        return True
    for bi in type_.recursive_bases:
        assert isinstance(bi, class_declaration.hierarchy_info_t)
        if bi.access_type != class_declaration.ACCESS_TYPES.PUBLIC:
            continue
        operators = bi.related_class.member_operators(
            function=matchers.custom_matcher_t(
                lambda decl: not decl.is_artificial) &
            matchers.access_type_matcher_t('public'),
            symbol=operator_symbol, allow_empty=True, recursive=False)
        if operators:
            return True
    return False


def has_public_equal(decl_type):
    """returns True, if class has public operator==, otherwise False"""
    return has_public_binary_operator(decl_type, '==')


def has_public_less(decl_type):
    """returns True, if class has public operator<, otherwise False"""
    return has_public_binary_operator(decl_type, '<')



# Copyright 2014-2017 Insight Software Consortium.
# Copyright 2004-2009 Roman Yakovenko.
# Distributed under the Boost Software License, Version 1.0.
# See http://www.boost.org/LICENSE_1_0.txt

import os
import fnmatch
import subprocess


def test_example():
    """Runs the example in the docs directory"""

    env = os.environ.copy()

    # Get the path to current directory
    path = os.path.dirname(os.path.realpath(__file__))
    # Set the COVERAGE_PROCESS_START env. variable.
    # Allows to cover files run in a subprocess
    # http://nedbatchelder.com/code/coverage/subprocess.html
    env["COVERAGE_PROCESS_START"] = path + "/../.coveragerc"

    # Find all the examples files
    file_paths = []
    for root, _, filenames in os.walk(path + "/../docs/examples"):
        for file_path in fnmatch.filter(filenames, '*.py'):
            file_paths.append(os.path.join(root, file_path))

    for file_path in file_paths:
        return_code = subprocess.call(
            ["python", path + "/example_tester_wrap.py", file_path],
            env=env)
        assert return_code == 0



# Copyright 2014-2017 Insight Software Consortium.
# Copyright 2004-2009 Roman Yakovenko.
# Distributed under the Boost Software License, Version 1.0.
# See http://www.boost.org/LICENSE_1_0.txt

import os

from . import autoconfig

from pygccxml import parser
from pygccxml import declarations


TEST_FILES = [
    "declarations_enums.hpp",
]


def test_cache():
    cache_file = os.path.join(autoconfig.data_directory, 'pygccxml.cache')
    if os.path.exists(cache_file) and os.path.isfile(cache_file):
        os.remove(cache_file)

    config = autoconfig.cxx_parsers_cfg.config.clone()

    cache = parser.file_cache_t(cache_file)
    reader = parser.source_reader_t(config, cache)
    decls1 = reader.read_file(TEST_FILES[0])
    cache.flush()
    cache = parser.file_cache_t(cache_file)
    reader = parser.source_reader_t(config, cache)
    decls2 = reader.read_file(TEST_FILES[0])

    enum_matcher = declarations.declaration_matcher_t(
        name="EColor",
        decl_type=declarations.enumeration_t
    )

    color1 = declarations.matcher.get_single(enum_matcher, decls1)
    color2 = declarations.matcher.get_single(enum_matcher, decls2)
    assert color1.values == color2.values



# Copyright 2014-2017 Insight Software Consortium.
# Copyright 2004-2009 Roman Yakovenko.
# Distributed under the Boost Software License, Version 1.0.
# See http://www.boost.org/LICENSE_1_0.txt

import os

from . import autoconfig

from pygccxml import parser
from pygccxml import declarations

TEST_FILE = "core_types.hpp"


def test_read_xml_file():
    config = autoconfig.cxx_parsers_cfg.config.clone()

    src_reader = parser.source_reader_t(config)
    src_decls = src_reader.read_file(TEST_FILE)

    xmlfile = src_reader.create_xml_file(TEST_FILE)

    conf_t = parser.file_configuration_t
    fconfig = conf_t(
        data=xmlfile,
        start_with_declarations=None,
        content_type=conf_t.CONTENT_TYPE.GCCXML_GENERATED_FILE)

    prj_reader = parser.project_reader_t(config)
    prj_decls = prj_reader.read_files(
        [fconfig],
        compilation_mode=parser.COMPILATION_MODE.FILE_BY_FILE)

    declarations.dump_declarations(
        src_decls,
        os.path.join(
            autoconfig.build_directory,
            'xmlfile_reader.src.txt'))
    declarations.dump_declarations(
        prj_decls,
        os.path.join(
            autoconfig.build_directory,
            'xmlfile_reader.prj.txt'))

    assert src_decls == prj_decls



# Copyright 2014-2017 Insight Software Consortium.
# Copyright 2004-2009 Roman Yakovenko.
# Distributed under the Boost Software License, Version 1.0.
# See http://www.boost.org/LICENSE_1_0.txt

"""
defines classes, that describe C++ types
"""

from . import algorithms_cache
from . import byte_info
from . import elaborated_info


class type_t(byte_info.byte_info):

    """base class for all types"""

    def __init__(self):
        byte_info.byte_info.__init__(self)
        self.cache = algorithms_cache.type_algs_cache_t()

    def __str__(self):
        res = self.decl_string
        if res[:2] == "::":
            res = res[2:]
        return res

    def __eq__(self, other):
        if not isinstance(other, type_t):
            return False
        return self.decl_string == other.decl_string

    def __hash__(self):
        return hash(self.decl_string)

    def __ne__(self, other):
        return not self.__eq__(other)

    def __lt__(self, other):
        if not isinstance(other, self.__class__):
            return self.__class__.__name__ < other.__class__.__name__
        return self.decl_string < other.decl_string

    def build_decl_string(self, with_defaults=True):
        raise NotImplementedError()

    @property
    def decl_string(self):
        if self.cache.decl_string is None:
            self.cache.decl_string = self.build_decl_string()
        return self.cache.decl_string

    @property
    def partial_decl_string(self):
        if self.cache.partial_decl_string is None:
            self.cache.partial_decl_string = self.build_decl_string(False)
        return self.cache.partial_decl_string

    def _clone_impl(self):
        raise NotImplementedError()

    def clone(self):
        """returns new instance of the type"""
        answer = self._clone_impl()
        return answer


# There are cases when GCC-XML reports something like this
# <Unimplemented id="_9482" tree_code="188" \
# tree_code_name="template_type_parm" node="0xcc4d5b0"/>
# In this case I will use this as type


class dummy_type_t(type_t):

    """provides :class:`type_t` interface for a string, that defines C++ type.

    This class could be very useful in the code generator.
    """

    def __init__(self, decl_string):
        type_t.__init__(self)
        self._decl_string = decl_string

    def build_decl_string(self, with_defaults=True):
        return self._decl_string

    def _clone_impl(self):
        return dummy_type_t(self._decl_string)


class unknown_t(type_t):
    """
    type, that represents all C++ types,
    that could not be parsed  by GCC-XML

    """

    def __init__(self):
        type_t.__init__(self)

    def build_decl_string(self, with_defaults=True):
        return '?unknown?'

    def _clone_impl(self):
        return self


class ellipsis_t(type_t):

    """type, that represents "..." in function definition"""

    def __init__(self):
        type_t.__init__(self)

    def build_decl_string(self, with_defaults=True):
        return '...'

    def _clone_impl(self):
        return self

##########################################################################
# Fundamental types:


class fundamental_t(type_t):

    """base class for all fundamental, build-in types"""

    def __init__(self, name):
        type_t.__init__(self)
        self._name = name

    def build_decl_string(self, with_defaults=True):
        return self._name

    def _clone_impl(self):
        return self


class java_fundamental_t(fundamental_t):

    """base class for all JNI defined fundamental types"""

    def __init__(self, name):
        fundamental_t.__init__(self, name)


class void_t(fundamental_t):

    """represents void type"""
    CPPNAME = 'void'

    def __init__(self):
        fundamental_t.__init__(self, void_t.CPPNAME)


class char_t(fundamental_t):

    """represents char type"""
    CPPNAME = 'char'

    def __init__(self):
        fundamental_t.__init__(self, char_t.CPPNAME)


class signed_char_t(fundamental_t):

    """represents signed char type"""
    CPPNAME = 'signed char'

    def __init__(self):
        fundamental_t.__init__(self, signed_char_t.CPPNAME)


class unsigned_char_t(fundamental_t):

    """represents unsigned char type"""
    CPPNAME = 'unsigned char'

    def __init__(self):
        fundamental_t.__init__(self, unsigned_char_t.CPPNAME)


class wchar_t(fundamental_t):

    """represents wchar_t type"""
    CPPNAME = 'wchar_t'

    def __init__(self):
        fundamental_t.__init__(self, wchar_t.CPPNAME)


class short_int_t(fundamental_t):

    """represents short int type"""
    CPPNAME = 'short int'

    def __init__(self):
        fundamental_t.__init__(self, short_int_t.CPPNAME)


class short_unsigned_int_t(fundamental_t):

    """represents short unsigned int type"""
    CPPNAME = 'short unsigned int'

    def __init__(self):
        fundamental_t.__init__(self, short_unsigned_int_t.CPPNAME)


class bool_t(fundamental_t):

    """represents bool type"""
    CPPNAME = 'bool'

    def __init__(self):
        fundamental_t.__init__(self, bool_t.CPPNAME)


class int_t(fundamental_t):

    """represents int type"""
    CPPNAME = 'int'

    def __init__(self):
        fundamental_t.__init__(self, int_t.CPPNAME)


class unsigned_int_t(fundamental_t):

    """represents unsigned int type"""
    CPPNAME = 'unsigned int'

    def __init__(self):
        fundamental_t.__init__(self, unsigned_int_t.CPPNAME)


class long_int_t(fundamental_t):

    """represents long int type"""
    CPPNAME = 'long int'

    def __init__(self):
        fundamental_t.__init__(self, long_int_t.CPPNAME)


class long_unsigned_int_t(fundamental_t):

    """represents long unsigned int type"""
    CPPNAME = 'long unsigned int'

    def __init__(self):
        fundamental_t.__init__(self, long_unsigned_int_t.CPPNAME)


class long_long_int_t(fundamental_t):

    """represents long long int type"""
    CPPNAME = 'long long int'

    def __init__(self):
        fundamental_t.__init__(self, long_long_int_t.CPPNAME)


class long_long_unsigned_int_t(fundamental_t):

    """represents long long unsigned int type"""
    CPPNAME = 'long long unsigned int'

    def __init__(self):
        fundamental_t.__init__(self, long_long_unsigned_int_t.CPPNAME)


class float_t(fundamental_t):

    """represents float type"""
    CPPNAME = 'float'

    def __init__(self):
        fundamental_t.__init__(self, float_t.CPPNAME)


class double_t(fundamental_t):

    """represents double type"""
    CPPNAME = 'double'

    def __init__(self):
        fundamental_t.__init__(self, double_t.CPPNAME)


class long_double_t(fundamental_t):

    """represents long double type"""
    CPPNAME = 'long double'

    def __init__(self):
        fundamental_t.__init__(self, long_double_t.CPPNAME)


class complex_double_t(fundamental_t):

    """represents complex double type"""
    CPPNAME = 'complex double'

    def __init__(self):
        fundamental_t.__init__(self, complex_double_t.CPPNAME)


class complex_long_double_t(fundamental_t):

    """represents complex long double type"""
    CPPNAME = 'complex long double'

    def __init__(self):
        fundamental_t.__init__(self, complex_long_double_t.CPPNAME)


class complex_float_t(fundamental_t):

    """represents complex float type"""
    CPPNAME = 'complex float'

    def __init__(self):
        fundamental_t.__init__(self, complex_float_t.CPPNAME)


class jbyte_t(java_fundamental_t):

    """represents jbyte type"""
    JNAME = 'jbyte'

    def __init__(self):
        java_fundamental_t.__init__(self, jbyte_t.JNAME)


class jshort_t(java_fundamental_t):

    """represents jshort type"""
    JNAME = 'jshort'

    def __init__(self):
        java_fundamental_t.__init__(self, jshort_t.JNAME)


class jint_t(java_fundamental_t):

    """represents jint type"""
    JNAME = 'jint'

    def __init__(self):
        java_fundamental_t.__init__(self, jint_t.JNAME)


class jlong_t(java_fundamental_t):

    """represents jlong type"""
    JNAME = 'jlong'

    def __init__(self):
        java_fundamental_t.__init__(self, jlong_t.JNAME)


class jfloat_t(java_fundamental_t):

    """represents jfloat type"""
    JNAME = 'jfloat'

    def __init__(self):
        java_fundamental_t.__init__(self, jfloat_t.JNAME)


class jdouble_t(java_fundamental_t):

    """represents jdouble type"""
    JNAME = 'jdouble'

    def __init__(self):
        java_fundamental_t.__init__(self, jdouble_t.JNAME)


class jchar_t(java_fundamental_t):

    """represents jchar type"""
    JNAME = 'jchar'

    def __init__(self):
        java_fundamental_t.__init__(self, jchar_t.JNAME)


class jboolean_t(java_fundamental_t):

    """represents jboolean type"""
    JNAME = 'jboolean'

    def __init__(self):
        java_fundamental_t.__init__(self, jboolean_t.JNAME)


class int128_t(fundamental_t):

    """represents __int128_t type"""
    CPPNAME = '__int128_t'

    def __init__(self):
        fundamental_t.__init__(self, int128_t.CPPNAME)


class uint128_t(fundamental_t):

    """represents __uint128_t type"""
    CPPNAME = '__uint128_t'

    def __init__(self):
        fundamental_t.__init__(self, uint128_t.CPPNAME)


FUNDAMENTAL_TYPES = {
    # adding java types
    void_t.CPPNAME: void_t(),
    char_t.CPPNAME: char_t(),
    signed_char_t.CPPNAME: signed_char_t(),
    unsigned_char_t.CPPNAME: unsigned_char_t(),
    wchar_t.CPPNAME: wchar_t(),
    short_int_t.CPPNAME: short_int_t(),
    'signed ' + short_int_t.CPPNAME: short_int_t(),
    short_unsigned_int_t.CPPNAME: short_unsigned_int_t(),
    bool_t.CPPNAME: bool_t(),
    int_t.CPPNAME: int_t(),
    'signed ' + int_t.CPPNAME: int_t(),
    unsigned_int_t.CPPNAME: unsigned_int_t(),
    long_int_t.CPPNAME: long_int_t(),
    long_unsigned_int_t.CPPNAME: long_unsigned_int_t(),
    long_long_int_t.CPPNAME: long_long_int_t(),
    long_long_unsigned_int_t.CPPNAME: long_long_unsigned_int_t(),
    int128_t.CPPNAME: int128_t(),
    uint128_t.CPPNAME: uint128_t(),
    float_t.CPPNAME: float_t(),
    double_t.CPPNAME: double_t(),
    long_double_t.CPPNAME: long_double_t(),
    complex_long_double_t.CPPNAME: complex_long_double_t(),
    complex_double_t.CPPNAME: complex_double_t(),
    complex_float_t.CPPNAME: complex_float_t(),
    jbyte_t.JNAME: jbyte_t(),
    jshort_t.JNAME: jshort_t(),
    jint_t.JNAME: jint_t(),
    jlong_t.JNAME: jlong_t(),
    jfloat_t.JNAME: jfloat_t(),
    jdouble_t.JNAME: jdouble_t(),
    jchar_t.JNAME: jchar_t(),
    jboolean_t.JNAME: jboolean_t(),
    '__java_byte': jbyte_t(),
    '__java_short': jshort_t(),
    '__java_int': jint_t(),
    '__java_long': jlong_t(),
    '__java_float': jfloat_t(),
    '__java_double': jdouble_t(),
    '__java_char': jchar_t(),
    '__java_boolean': jboolean_t()
}
"""
defines a mapping between fundamental type name and its synonym to the instance
of class that describes the type
"""


def _f(x, with_defaults):
    """
    A small helper function.

    """

    return x.build_decl_string(with_defaults)

##########################################################################
# Compound types:


class compound_t(type_t):

    """class that allows to represent compound types like `const int*`"""

    def __init__(self, base):
        type_t.__init__(self)
        self._base = base

    @property
    def base(self):
        """reference to internal/base class"""
        return self._base

    @base.setter
    def base(self, new_base):
        self._base = new_base

    def build_decl_string(self, with_defaults=True):
        raise NotImplementedError()

    def _clone_impl(self):
        raise NotImplementedError()


class volatile_t(compound_t):

    """represents `volatile whatever` type"""

    def __init__(self, base):
        compound_t.__init__(self, base)

    def build_decl_string(self, with_defaults=True):
        return self.base.build_decl_string(with_defaults) + ' volatile'

    def _clone_impl(self):
        return volatile_t(self.base.clone())


class restrict_t(compound_t):

    """represents `restrict whatever` type"""

    # The restrict keyword can be considered an extension to the strict
    # aliasing rule. It allows the programmer to declare that pointers which
    # share the same type (or were otherwise validly created) do not alias
    # eachother. By using restrict the programmer can declare that any loads
    # and stores through the qualified pointer (or through another pointer
    # copied either directly or indirectly from the restricted pointer) are
    # the only loads and stores to the same address during the lifetime of
    # the pointer. In other words, the pointer is not aliased by any pointers
    # other than its own copies.

    def __init__(self, base):
        compound_t.__init__(self, base)

    def build_decl_string(self, with_defaults=True):
        return '__restrict__ ' + self.base.build_decl_string(with_defaults)

    def _clone_impl(self):
        return restrict_t(self.base.clone())


class const_t(compound_t):

    """represents `whatever const` type"""

    def __init__(self, base):
        compound_t.__init__(self, base)

    def build_decl_string(self, with_defaults=True):
        return self.base.build_decl_string(with_defaults) + ' const'

    def _clone_impl(self):
        return const_t(self.base.clone())


class pointer_t(compound_t):

    """represents `whatever*` type"""

    def __init__(self, base):
        compound_t.__init__(self, base)

    def build_decl_string(self, with_defaults=True):
        decl_string = self.base.build_decl_string(with_defaults)
        if isinstance(self.base, calldef_type_t):
            # This is a function pointer. Do not add supplementary *
            return decl_string
        return decl_string + " *"

    def _clone_impl(self):
        return pointer_t(self.base.clone())


class reference_t(compound_t):

    """represents `whatever&` type"""

    def __init__(self, base):
        compound_t.__init__(self, base)

    def build_decl_string(self, with_defaults=True):
        return self.base.build_decl_string(with_defaults) + ' &'

    def _clone_impl(self):
        return reference_t(self.base.clone())


class elaborated_t(compound_t):

    """represents `elaborated` type"""

    def __init__(self, base):
        compound_t.__init__(self, base)

    def build_decl_string(self, with_defaults=True):
        prefix = ""
        if isinstance(self.base, type(declarated_t)) and \
                isinstance(self.base.declaration, type(elaborated_info)):
            prefix = self.base.declaration.elaborated_type_specifier + " "
        return prefix + self.base.build_decl_string(with_defaults)

    def _clone_impl(self):
        return elaborated_t(self.base.clone())


class array_t(compound_t):

    """represents C++ array type"""
    SIZE_UNKNOWN = -1

    def __init__(self, base, size):
        compound_t.__init__(self, base)
        self._size = size

    @property
    def size(self):
        """returns array size"""
        return self._size

    @size.setter
    def size(self, size):
        """sometimes there is a need to update the size of the array"""
        self.cache.reset()
        self._size = size

    def build_decl_string(self, with_defaults=True):
        return self.__bds_for_multi_dim_arrays(None, with_defaults)

    def __bds_for_multi_dim_arrays(self, parent_dims=None, with_defaults=True):
        if parent_dims:
            parent_dims.append(self.size)
        else:
            parent_dims = [self.size]

        if isinstance(self.base, array_t):
            return self.base.__bds_for_multi_dim_arrays(
                parent_dims,
                with_defaults)

        tmp = []
        for s in parent_dims:
            tmp.append('[%d]' % s)
        return self.base.build_decl_string(with_defaults) + " " + "".join(tmp)

    def _clone_impl(self):
        return array_t(self.base.clone(), self.size)


class calldef_type_t(object):

    """base class for all types that describes "callable" declaration"""

    def __init__(self, return_type=None, arguments_types=None):
        object.__init__(self)
        self._return_type = return_type
        if arguments_types is None:
            arguments_types = []
        self._arguments_types = arguments_types

    @property
    def return_type(self):
        """reference to :class:`return type <type_t>`"""
        return self._return_type

    @return_type.setter
    def return_type(self, new_return_type):
        self._return_type = new_return_type

    @property
    def arguments_types(self):
        """list of argument :class:`types <type_t>`"""
        return self._arguments_types

    @arguments_types.setter
    def arguments_types(self, new_arguments_types):
        self._arguments_types = new_arguments_types

    @property
    def has_ellipsis(self):
        return self.arguments_types and isinstance(
            self.arguments_types[-1], ellipsis_t)


class free_function_type_t(type_t, calldef_type_t):

    """describes free function type"""
    NAME_TEMPLATE = '%(return_type)s (*)( %(arguments)s )'
    TYPEDEF_NAME_TEMPLATE = (
        '%(return_type)s ( *%(typedef_name)s )( %(arguments)s )')

    def __init__(self, return_type=None, arguments_types=None):
        type_t.__init__(self)
        calldef_type_t.__init__(self, return_type, arguments_types)

    @staticmethod
    def create_decl_string(
            return_type, arguments_types, with_defaults=True):
        """
        Returns free function type

        :param return_type: function return type
        :type return_type: :class:`type_t`
        :param arguments_types: list of argument :class:`type <type_t>`
        :rtype: :class:`free_function_type_t`

        """

        return free_function_type_t.NAME_TEMPLATE % {
            'return_type': return_type.build_decl_string(with_defaults),
            'arguments': ','.join(
                [_f(x, with_defaults) for x in arguments_types])}

    def build_decl_string(self, with_defaults=True):
        return self.create_decl_string(
            self.return_type,
            self.arguments_types,
            with_defaults)

    def _clone_impl(self):
        rt_clone = None
        if self.return_type:
            rt_clone = self.return_type.clone()
        return free_function_type_t(
            rt_clone, [
                arg.clone() for arg in self.arguments_types])

    # TODO: create real typedef
    def create_typedef(self, typedef_name, unused=None, with_defaults=True):
        """returns string, that contains valid C++ code, that defines typedef
        to function type

        :param name: the desired name of typedef
        """

        return free_function_type_t.TYPEDEF_NAME_TEMPLATE % {
            'typedef_name': typedef_name,
            'return_type': self.return_type.build_decl_string(with_defaults),
            'arguments': ','.join(
                [_f(x, with_defaults) for x in self.arguments_types])}


class member_function_type_t(type_t, calldef_type_t):

    """describes member function type"""
    NAME_TEMPLATE = (
        '%(return_type)s ( %(class)s::* )( %(arguments)s )%(has_const)s')
    TYPEDEF_NAME_TEMPLATE = (
        '%(return_type)s ( %(class)s::*%(typedef_name)s' +
        ')( %(arguments)s ) %(has_const)s')

    def __init__(
            self,
            class_inst=None,
            return_type=None,
            arguments_types=None,
            has_const=False):
        type_t.__init__(self)
        calldef_type_t.__init__(self, return_type, arguments_types)
        self._has_const = has_const
        self._class_inst = class_inst

    @property
    def has_const(self):
        """describes, whether function has const modifier"""
        return self._has_const

    @has_const.setter
    def has_const(self, has_const):
        self._has_const = has_const

    @property
    def class_inst(self):
        """reference to parent :class:`class <declaration_t>`"""
        return self._class_inst

    @class_inst.setter
    def class_inst(self, class_inst):
        self._class_inst = class_inst

    # TODO: create real typedef
    def create_typedef(
            self,
            typedef_name,
            class_alias=None,
            with_defaults=True):
        """creates typedef to the function type

        :param typedef_name: desired type name
        :rtype: string
        """
        has_const_str = ''
        if self.has_const:
            has_const_str = 'const'
        if None is class_alias:
            if with_defaults:
                class_alias = self.class_inst.decl_string
            else:
                class_alias = self.class_inst.partial_decl_string

        return member_function_type_t.TYPEDEF_NAME_TEMPLATE % {
            'typedef_name': typedef_name,
            'return_type': self.return_type.build_decl_string(with_defaults),
            'class': class_alias,
            'arguments': ','.join(
                [_f(x, with_defaults) for x in self.arguments_types]),
            'has_const': has_const_str}

    @staticmethod
    def create_decl_string(
            return_type,
            class_decl_string,
            arguments_types,
            has_const,
            with_defaults=True):
        has_const_str = ''
        if has_const:
            has_const_str = 'const'
        return_type_decl_string = ''
        if return_type:
            return_type_decl_string = return_type.build_decl_string(
                with_defaults)

        return member_function_type_t.NAME_TEMPLATE % {
            'return_type': return_type_decl_string,
            'class': class_decl_string,
            'arguments': ','.join(
                [_f(x, with_defaults) for x in arguments_types]),
            'has_const': has_const_str}

    def build_decl_string(self, with_defaults=True):
        return self.create_decl_string(
            self.return_type,
            self.class_inst.decl_string,
            self.arguments_types,
            self.has_const,
            with_defaults)

    def _clone_impl(self):
        rt_clone = None
        if self.return_type:
            rt_clone = self.return_type.clone()

        return member_function_type_t(
            self.class_inst, rt_clone, [
                arg.clone() for arg in self.arguments_types], self.has_const)


class member_variable_type_t(compound_t):

    """describes member variable type"""
    NAME_TEMPLATE = '%(type)s ( %(class)s::* )'

    def __init__(self, class_inst=None, variable_type=None):
        compound_t.__init__(self, class_inst)
        self._mv_type = variable_type

    @property
    def variable_type(self):
        """describes member variable :class:`type <type_t>`"""
        return self._mv_type

    @variable_type.setter
    def variable_type(self, new_type):
        self._mv_type = new_type

    def build_decl_string(self, with_defaults=True):
        return self.NAME_TEMPLATE % {
            'type': self.variable_type.build_decl_string(with_defaults),
            'class': self.base.build_decl_string(with_defaults)}

    def _clone_impl(self):
        return member_variable_type_t(
            class_inst=self.base,
            variable_type=self.variable_type.clone())


##########################################################################
# declarated types:

class declarated_t(type_t, byte_info.byte_info):

    """class that binds between to hierarchies: :class:`type_t`
    and :class:`declaration_t`"""

    def __init__(self, declaration):
        type_t.__init__(self)
        byte_info.byte_info.__init__(self)
        self._declaration = declaration
        self.byte_size = self._declaration.byte_size
        self.byte_align = self._declaration.byte_align

    @property
    def declaration(self):
        """reference to :class:`declaration_t`"""
        return self._declaration

    @declaration.setter
    def declaration(self, new_declaration):
        self._declaration = new_declaration
        self.byte_size = self._declaration.byte_size
        self.byte_align = self._declaration.byte_align

    def build_decl_string(self, with_defaults=True):
        if with_defaults:
            return self._declaration.decl_string

        return self._declaration.partial_decl_string

    def _clone_impl(self):
        return declarated_t(self._declaration)


class type_qualifiers_t(object):

    """contains additional information about type: mutable, static, extern"""

    def __init__(self, has_static=False, has_mutable=False, has_extern=False):
        self._has_static = has_static
        self._has_extern = has_extern
        self._has_mutable = has_mutable

    def __eq__(self, other):
        if not isinstance(other, type_qualifiers_t):
            return False
        return self.has_static == other.has_static \
            and self.has_extern == other.has_extern \
            and self.has_mutable == other.has_mutable

    __hash__ = object.__hash__

    def __ne__(self, other):
        return not self.__eq__(other)

    def __lt__(self, other):
        if not isinstance(other, type_qualifiers_t):
            return object.__lt__(self, other)
        return self.has_static < other.has_static \
            and self.has_extern < other.has_extern \
            and self.has_mutable < other.has_mutable

    @property
    def has_static(self):
        return self._has_static

    @has_static.setter
    def has_static(self, has_static):
        self._has_static = has_static

    @property
    def has_extern(self):
        return self._has_extern

    @has_extern.setter
    def has_extern(self, has_extern):
        self._has_extern = has_extern

    @property
    def has_mutable(self):
        return self._has_mutable

    @has_mutable.setter
    def has_mutable(self, has_mutable):
        self._has_mutable = has_mutable



# Copyright 2014-2017 Insight Software Consortium.
# Copyright 2004-2009 Roman Yakovenko.
# Distributed under the Boost Software License, Version 1.0.
# See http://www.boost.org/LICENSE_1_0.txt

import os

import pytest

from pygccxml import parser
from pygccxml import utils


def test_config_warn():
    """
        Test that a missing include directory is printing a warning,
        not raising an error
    """

    # Some code to parse for the example
    code = "int a;"

    # Find the location of the xml generator (castxml or gccxml)
    generator_path, name = utils.find_xml_generator()

    # Path given as include director doesn't exist
    config = parser.xml_generator_configuration_t(
        xml_generator_path=generator_path,
        xml_generator=name,
        include_paths=["doesnt/exist", os.getcwd()])
    with pytest.warns(RuntimeWarning):
        parser.parse_string(code, config)



# Copyright 2014-2021 Insight Software Consortium.
# Copyright 2004-2009 Roman Yakovenko.
# Distributed under the Boost Software License, Version 1.0.
# See http://www.boost.org/LICENSE_1_0.txt

import pytest

from . import autoconfig

from pygccxml import parser
from pygccxml import declarations


TEST_FILES = [
    "test_ccflags.hpp",
]

COMPILATION_MODE = parser.COMPILATION_MODE.ALL_AT_ONCE


@pytest.fixture
def config():
    config = autoconfig.cxx_parsers_cfg.config.clone()
    config.castxml_epic_version = 1
    config.append_cflags("-fopenmp")
    return config


def test_ccflags(config):
    # First check that macro is not defined.
    decls = parser.parse(TEST_FILES, config, COMPILATION_MODE)
    global_ns = declarations.get_global_namespace(decls)

    namespace_names = [
        n.name for n in global_ns.namespaces(allow_empty=True)
    ]
    assert "ccflags_test_namespace" not in namespace_names

    # Next check that macro is defined when passed directly as ccflag

    if "clang++" in config.compiler_path:
        config.append_ccflags("-Xpreprocessor")
    config.append_ccflags("-fopenmp")

    decls = parser.parse(TEST_FILES, config, COMPILATION_MODE)
    global_ns = declarations.get_global_namespace(decls)

    namespace_names = [n.name for n in global_ns.namespaces()]
    assert "ccflags_test_namespace" in namespace_names



# Copyright 2014-2017 Insight Software Consortium.
# Copyright 2004-2009 Roman Yakovenko.
# Distributed under the Boost Software License, Version 1.0.
# See http://www.boost.org/LICENSE_1_0.txt

import os
import bz2

from . import autoconfig

from pygccxml import parser
from pygccxml import declarations


def test_copy_constructor2():
    # Extract the xml file from the bz2 archive
    bz2_path = os.path.join(
        autoconfig.data_directory,
        'ogre.1.7.xml.bz2')
    xml_path = os.path.join(
        autoconfig.data_directory,
        'ogre.1.7.xml')
    with open(xml_path, 'wb') as new_file:
        # bz2.BZ2File can not be used in a with statement in python 2.6
        bz2_file = bz2.BZ2File(bz2_path, 'rb')
        for data in iter(lambda: bz2_file.read(100 * 1024), b''):
            new_file.write(data)
        bz2_file.close()

    reader = parser.source_reader_t(autoconfig.cxx_parsers_cfg.config)
    global_ns = declarations.get_global_namespace(
        reader.read_xml_file(xml_path)
        )
    global_ns.init_optimizer()

    for x in global_ns.typedefs('SettingsMultiMap'):
        assert declarations.is_noncopyable(x) is False

    for x in global_ns.typedefs('SettingsIterator'):
        assert declarations.is_noncopyable(x) is False

    for x in global_ns.typedefs('SectionIterator'):
        assert declarations.is_noncopyable(x) is False

    os.remove(xml_path)



# Copyright 2014-2017 Insight Software Consortium.
# Copyright 2004-2009 Roman Yakovenko.
# Distributed under the Boost Software License, Version 1.0.
# See http://www.boost.org/LICENSE_1_0.txt

from pygccxml import utils
from pygccxml import declarations
from pygccxml import parser

import os
import sys
import warnings
warnings.simplefilter("error", Warning)
# Find out the file location within the sources tree
this_module_dir_path = os.path.abspath(
    os.path.dirname(sys.modules[__name__].__file__))

# Find out the c++ parser
generator_path, generator_name = utils.find_xml_generator()

# Configure the xml generator
xml_generator_config = parser.xml_generator_configuration_t(
    xml_generator_path=generator_path,
    xml_generator=generator_name)

# The c++ file we want to parse
filename = "example.hpp"
filename = this_module_dir_path + "/" + filename

decls = parser.parse([filename], xml_generator_config)
global_namespace = declarations.get_global_namespace(decls)
ns = global_namespace.namespace("ns")

# The variables() method will return a list of variables.
# We know that the c variable is the third one in the list:
c = ns.variables()[2]
print("My name is: " + c.name)
print("My type is: " + str(c.decl_type))
print("My value is: " + c.value)

# Of course you can also loop over the list and look for the right name
for var in ns.variables():
    if var.name == "c":
        print("My name is: " + var.name)
        print("My type is: " + str(var.decl_type))
        print("My value is: " + var.value)

# One way to get a variable is to use the variable() method and
# a lambda function. This is the most flexible way as you can implement
# your own lambda function to filter out variables following your
# specific criteria.
c = ns.variable(lambda v: v.name == "c")
print("My name is: " + c.name)
print("My type is: " + str(c.decl_type))
print("My value is: " + c.value)



# Copyright 2014-2017 Insight Software Consortium.
# Copyright 2004-2009 Roman Yakovenko.
# Distributed under the Boost Software License, Version 1.0.
# See http://www.boost.org/LICENSE_1_0.txt

import os
import sys
import logging
import warnings
import platform

# Prevents copy.deepcopy RecursionError in some tests (Travis build)
sys.setrecursionlimit(10000)

this_module_dir_path = os.path.abspath(
    os.path.dirname(sys.modules[__name__].__file__))

data_directory = os.path.join(this_module_dir_path, 'data')
build_directory = os.path.join(this_module_dir_path, 'temp')

if not os.path.exists(build_directory):
    os.makedirs(build_directory)

sys.path.insert(1, os.path.join(os.curdir, '..'))
# The tests are run on the parent pygccxml directory, not the one
# in site-packages. Insert the directory's path.
sys.path.insert(1, "../src/pygccxml")

from pygccxml import parser  # nopep8
from pygccxml import utils  # nopep8

# We want to make sure we throw an error for ALL the warnings during the
# tests. This will allow us to be notified by the build bots, so that the
# warnings can be fixed.
warnings.simplefilter("error", Warning)

# Set logging level
utils.loggers.set_level(logging.CRITICAL)

# Find out the c++ parser (gccxml or castxml)
generator_path, generator_name = utils.find_xml_generator()


class cxx_parsers_cfg(object):
    config = parser.load_xml_generator_configuration(
        os.path.normpath(this_module_dir_path + '/xml_generator.cfg'),
        xml_generator_path=generator_path,
        working_directory=data_directory,
        xml_generator=generator_name)

    if platform.system() == 'Windows':
        config.define_symbols.append('_HAS_EXCEPTIONS=0')


if cxx_parsers_cfg.config.xml_generator:
    generator_name = cxx_parsers_cfg.config.xml_generator
if cxx_parsers_cfg.config.xml_generator_path:
    generator_path = cxx_parsers_cfg.config.xml_generator_path



# Copyright 2014-2017 Insight Software Consortium.
# Copyright 2004-2009 Roman Yakovenko.
# Distributed under the Boost Software License, Version 1.0.
# See http://www.boost.org/LICENSE_1_0.txt

import os

from . import class_declaration
from . import type_traits
from . import enumeration
from . import calldef_members
from . import calldef_types
from . import scopedef
from . import cpptypes
from .. import utils


def is_union(declaration):
    """
    Returns True if declaration represents a C++ union

    Args:
        declaration (declaration_t): the declaration to be checked.

    Returns:
        bool: True if declaration represents a C++ union
    """
    if not is_class(declaration):
        return False
    decl = class_traits.get_declaration(declaration)
    return decl.class_type == class_declaration.CLASS_TYPES.UNION


def is_struct(declaration):
    """
    Returns True if declaration represents a C++ struct

    Args:
        declaration (declaration_t): the declaration to be checked.

    Returns:
        bool: True if declaration represents a C++ struct
    """
    if not is_class(declaration):
        return False
    decl = class_traits.get_declaration(declaration)
    return decl.class_type == class_declaration.CLASS_TYPES.STRUCT


class declaration_xxx_traits(object):
    """this class implements the functionality needed for convenient work with
    declaration classes

    Implemented functionality:
        - find out whether a declaration is a desired one
        - get reference to the declaration
    """

    def __init__(self, declaration_class):
        self.declaration_class = declaration_class

    @staticmethod
    def __apply_sequence(type_):
        return \
            type_traits.remove_declarated(
                type_traits.remove_elaborated(
                    type_traits.remove_cv(
                        type_traits.remove_alias(
                            type_traits.remove_pointer(type_)))))

    def is_my_case(self, type_):
        """returns True, if type represents the desired declaration,
        False otherwise"""
        return (
            isinstance(self.__apply_sequence(type_), self.declaration_class)
        )

    def get_declaration(self, type_):
        """returns reference to the declaration

        Precondition: self.is_my_case( type ) == True
        """
        return self.__apply_sequence(type_)


enum_traits = declaration_xxx_traits(enumeration.enumeration_t)
"""implements functionality, needed for convenient work with C++ enums"""

is_enum = enum_traits.is_my_case
"""returns True, if type represents C++ enumeration declaration,
False otherwise"""

enum_declaration = enum_traits.get_declaration
"""returns reference to enum declaration"""

class_traits = declaration_xxx_traits(class_declaration.class_t)
"""implements functionality, needed for convenient work with C++ classes"""

is_class = class_traits.is_my_case
"""returns True, if type represents C++ class definition, False otherwise"""

class_declaration_traits = declaration_xxx_traits(
    class_declaration.class_declaration_t)
"""implements functionality, needed for convenient work with C++ class
declarations"""

is_class_declaration = class_declaration_traits.is_my_case
"""returns True, if type represents C++ class declaration, False otherwise"""


def find_trivial_constructor(type_):
    """
    Returns reference to trivial constructor.

    Args:
        type_ (declarations.class_t): the class to be searched.

    Returns:
        declarations.constructor_t: the trivial constructor

    """
    assert isinstance(type_, class_declaration.class_t)

    trivial = type_.constructors(
        lambda x: is_trivial_constructor(x),
        recursive=False,
        allow_empty=True)
    if trivial:
        return trivial[0]

    return None


def find_copy_constructor(type_):
    """
    Returns reference to copy constructor.

    Args:
        type_ (declarations.class_t): the class to be searched.

    Returns:
        declarations.constructor_t: the copy constructor

    """
    copy_ = type_.constructors(
        lambda x: is_copy_constructor(x),
        recursive=False,
        allow_empty=True)
    if copy_:
        return copy_[0]

    return None


def find_noncopyable_vars(class_type, already_visited_cls_vars=None):
    """
    Returns list of all `noncopyable` variables.

    If an already_visited_cls_vars list is provided as argument, the returned
    list will not contain these variables. This list will be extended with
    whatever variables pointing to classes have been found.

    Args:
        class_type (declarations.class_t): the class to be searched.
        already_visited_cls_vars (list): optional list of vars that should not
            be checked a second time, to prevent infinite recursions.

    Returns:
        list: list of all `noncopyable` variables.

    """

    assert isinstance(class_type, class_declaration.class_t)

    logger = utils.loggers.cxx_parser
    mvars = class_type.variables(
        lambda v: not v.type_qualifiers.has_static,
        recursive=False,
        allow_empty=True)
    noncopyable_vars = []

    if already_visited_cls_vars is None:
        already_visited_cls_vars = []

    message = (
        "__contains_noncopyable_mem_var - %s - TRUE - " +
        "contains const member variable")

    for mvar in mvars:

        var_type = type_traits.remove_reference(mvar.decl_type)

        if type_traits.is_const(var_type):
            no_const = type_traits.remove_const(var_type)
            if type_traits.is_fundamental(no_const) or is_enum(no_const):
                logger.debug(
                    (message + "- fundamental or enum"),
                    var_type.decl_string)
                noncopyable_vars.append(mvar)
            if is_class(no_const):
                logger.debug((message + " - class"), var_type.decl_string)
                noncopyable_vars.append(mvar)
            if type_traits.is_array(no_const):
                logger.debug((message + " - array"), var_type.decl_string)
                noncopyable_vars.append(mvar)

        if type_traits.is_pointer(var_type):
            continue

        if class_traits.is_my_case(var_type):
            cls = class_traits.get_declaration(var_type)

            # Exclude classes that have already been visited.
            if cls in already_visited_cls_vars:
                continue
            already_visited_cls_vars.append(cls)

            if is_noncopyable(cls, already_visited_cls_vars):
                logger.debug(
                    (message + " - class that is not copyable"),
                    var_type.decl_string)
                noncopyable_vars.append(mvar)

    logger.debug((
        "__contains_noncopyable_mem_var - %s - FALSE - doesn't " +
        "contain noncopyable members"), class_type.decl_string)

    return noncopyable_vars


def has_trivial_constructor(class_):
    """if class has public trivial constructor, this function will return
    reference to it, None otherwise"""
    class_ = class_traits.get_declaration(class_)
    trivial = find_trivial_constructor(class_)
    if trivial and trivial.access_type == 'public':
        return trivial


def has_copy_constructor(class_):
    """if class has public copy constructor, this function will return
    reference to it, None otherwise"""
    class_ = class_traits.get_declaration(class_)
    copy_constructor = find_copy_constructor(class_)
    if copy_constructor and copy_constructor.access_type == 'public':
        return copy_constructor


def has_destructor(class_):
    """if class has destructor, this function will return reference to it,
    None otherwise"""
    class_ = class_traits.get_declaration(class_)
    destructor = class_.decls(
        decl_type=calldef_members.destructor_t,
        recursive=False,
        allow_empty=True)
    if destructor:
        return destructor[0]


def has_public_constructor(class_):
    """if class has any public constructor, this function will return list of
    them, otherwise None"""
    class_ = class_traits.get_declaration(class_)
    decls = class_.constructors(
        lambda c: not is_copy_constructor(c) and c.access_type == 'public',
        recursive=False,
        allow_empty=True)
    if decls:
        return decls


def has_public_assign(class_):
    """returns True, if class has public assign operator, False otherwise"""
    class_ = class_traits.get_declaration(class_)
    decls = class_.member_operators(
        lambda o: o.symbol == '=' and o.access_type == 'public',
        recursive=False,
        allow_empty=True)
    return bool(decls)


def has_public_destructor(decl_type):
    """returns True, if class has public destructor, False otherwise"""
    d = has_destructor(decl_type)
    return d and d.access_type == 'public'


def has_vtable(decl_type):
    """True, if class has virtual table, False otherwise"""
    assert isinstance(decl_type, class_declaration.class_t)
    return bool(
        decl_type.calldefs(
            lambda f: isinstance(f, calldef_members.member_function_t) and
            f.virtuality != calldef_types.VIRTUALITY_TYPES.NOT_VIRTUAL,
            recursive=False,
            allow_empty=True))


def is_base_and_derived(based, derived):
    """returns True, if there is "base and derived" relationship between
    classes, False otherwise"""
    assert isinstance(based, class_declaration.class_t)
    assert isinstance(derived, (class_declaration.class_t, tuple))

    if isinstance(derived, class_declaration.class_t):
        all_derived = ([derived])
    else:  # tuple
        all_derived = derived

    for derived_cls in all_derived:
        for base_desc in derived_cls.recursive_bases:
            if base_desc.related_class == based:
                return True
    return False


def has_any_non_copyconstructor(decl_type):
    """if class has any public constructor, which is not copy constructor,
    this function will return list of them, otherwise None"""
    class_ = class_traits.get_declaration(decl_type)
    decls = class_.constructors(
        lambda c: not is_copy_constructor(c) and c.access_type == 'public',
        recursive=False,
        allow_empty=True)
    if decls:
        return decls


class __is_convertible_t(object):

    """implementation details"""

    def __init__(self, source, target):
        self.__source = self.__normalize(source)
        self.__target = self.__normalize(target)

    @staticmethod
    def __find_class_by_class_declaration(class_decl):
        found = scopedef.find_declaration(
            class_decl.parent.declarations,
            name=class_decl.name,
            decl_type=class_declaration.class_t)
        return found

    def __normalize(self, type_):
        type_ = type_traits.remove_alias(type_)
        bt_of_type = type_traits.base_type(type_)
        if isinstance(bt_of_type, cpptypes.declarated_t) \
           and isinstance(bt_of_type.declaration,
                          class_declaration.class_declaration_t):
            type_ = type_.clone()
            bt_of_type = type_traits.base_type(type_)
            bt_of_type.declaration = self.__find_class_by_class_declaration(
                bt_of_type.declaration)
        return type_

    @staticmethod
    def __test_trivial(src, target):
        if not (src and target):
            return False
        if type_traits.is_same(src, target):
            return True  # X => X
        if type_traits.is_const(target) and type_traits.is_same(
                src, target.base):
            return True  # X => const X
        if type_traits.is_reference(target) and type_traits.is_same(
                src, target.base):
            return True  # X => X&
        if type_traits.is_reference(target) and type_traits.is_const(
                target.base) and type_traits.is_same(src, target.base.base):
            return True  # X => const X&
        if type_traits.is_same(target, cpptypes.pointer_t(cpptypes.void_t())):
            if type_traits.is_integral(src) or is_enum(src):
                return False
            return True  # X => void*
        if type_traits.is_pointer(src) and \
            type_traits.is_pointer(target) and \
            type_traits.is_const(target.base) and \
                type_traits.is_same(src.base, target.base.base):
            return True  # X* => const X*
        if type_traits.is_reference(src) and \
            type_traits.is_reference(target) and \
            type_traits.is_const(target.base) and \
                type_traits.is_same(src.base, target.base.base):
            return True  # X& => const X&
        if not type_traits.is_const(src) and \
            type_traits.is_array(src) and \
            type_traits.is_pointer(target) and \
                type_traits.is_same(type_traits.base_type(src), target.base):
            return True  # X[2] => X*
        if type_traits.is_array(src) and \
            type_traits.is_pointer(target) and \
            type_traits.is_const(target.base) and \
            type_traits.is_same(
                    type_traits.base_type(src), target.base.base):
            return True

    @staticmethod
    def __test_pointer_to_func_or_mv__to__func_or_mv(source, target):
        if type_traits.is_pointer(source) \
           and type_traits.is_reference(target) \
           and isinstance(target.base,
                          (cpptypes.free_function_type_t,
                           cpptypes.member_function_type_t,
                           cpptypes.member_variable_type_t)) \
           and type_traits.is_same(source.base, target.base):
            return True

        if type_traits.is_pointer(source) \
           and isinstance(target,
                          (cpptypes.free_function_type_t,
                           cpptypes.member_function_type_t,
                           cpptypes.member_variable_type_t)) \
           and type_traits.is_same(source.base, target):
            return True

        if type_traits.is_pointer(target) \
           and type_traits.is_reference(source) \
           and isinstance(source.base,
                          (cpptypes.free_function_type_t,
                           cpptypes.member_function_type_t,
                           cpptypes.member_variable_type_t)) \
           and type_traits.is_same(source.base, target.base):
            return True

        if type_traits.is_pointer(target) \
           and isinstance(source,
                          (cpptypes.free_function_type_t,
                           cpptypes.member_function_type_t,
                           cpptypes.member_variable_type_t)) \
           and type_traits.is_same(target.base, source):
            return True

    @staticmethod
    def __test_const_x_ref__to__x(source, target):
        if not type_traits.is_reference(source) \
           or not type_traits.is_const(source.base) \
           or not type_traits.is_same(source.base.base, target):
            return False
        if type_traits.is_fundamental(target):
            return True
        if is_enum(target):
            return True
        if isinstance(target, cpptypes.declarated_t):
            assert isinstance(target.declaration, class_declaration.class_t)
            if has_copy_constructor(target.declaration):
                return True  # we have copy constructor
        return False

    @staticmethod
    def __test_const_ref_x__to__y(source, target):
        if not type_traits.is_reference(source) or not \
                type_traits.is_const(source.base):
            return False
        if type_traits.is_fundamental(source.base.base) and \
                type_traits.is_fundamental(target):
            return True
        if is_convertible(source.base.base, cpptypes.int_t()) and \
                is_enum(target):
            return True
        if isinstance(target, cpptypes.declarated_t):
            assert isinstance(target.declaration, class_declaration.class_t)
            if has_copy_constructor(target.declaration):
                return True  # we have copy constructor
        return False

    @staticmethod
    def __test_ref_x__to__x(source, target):
        if not type_traits.is_reference(source) or not \
                type_traits.is_same(source.base, target):
            return False
        if type_traits.is_fundamental(target):
            return True
        if is_enum(target):
            return True
        if isinstance(target, cpptypes.declarated_t):
            assert isinstance(target.declaration, class_declaration.class_t)
            if has_copy_constructor(target.declaration):
                return True  # we have copy constructor
        return False

    @staticmethod
    def __test_ref_x__to__y(source, target):
        if not type_traits.is_reference(source):
            return False
        if type_traits.is_fundamental(source.base) and \
                type_traits.is_fundamental(target):
            return True
        if is_convertible(source.base, cpptypes.int_t()) and is_enum(target):
            return True
        if isinstance(target, cpptypes.declarated_t):
            assert isinstance(target.declaration, class_declaration.class_t)
            if has_copy_constructor(target.declaration):
                return True  # we have copy constructor
        return False

    @staticmethod
    def __test_fundamental__to__fundamental(source, target):
        if not type_traits.is_fundamental(
                type_traits.base_type(source)) or not \
                type_traits.is_fundamental(
                    type_traits.base_type(target)):
            return False
        if type_traits.is_void(type_traits.base_type(source)) or \
                type_traits.is_void(type_traits.base_type(target)):
            return False
        if type_traits.is_fundamental(source) and \
                type_traits.is_fundamental(target):
            return True
        if not type_traits.is_pointer(source) and \
                type_traits.is_fundamental(target):
            return True
        if not type_traits.is_pointer(source) and \
                type_traits.is_const(target) and \
                type_traits.is_fundamental(target.base):
            return True
        if type_traits.is_fundamental(source) \
           and type_traits.is_reference(target) \
           and type_traits.is_const(target.base) \
           and type_traits.is_fundamental(target.base.base):
            return True  # X => const Y&
        return False

    @staticmethod
    def _is_both_declarated(x, y):
        return (
            isinstance(x, cpptypes.declarated_t) and
            isinstance(y, cpptypes.declarated_t))

    def __test_derived_to_based(self, source, target):
        derived = type_traits.base_type(source)
        base = type_traits.base_type(target)
        if not (
                isinstance(derived, cpptypes.declarated_t) and
                isinstance(derived.declaration, class_declaration.class_t)):
            return False
        if not (isinstance(base, cpptypes.declarated_t) and
                isinstance(base.declaration, class_declaration.class_t)):
            return False
        base = base.declaration
        derived = derived.declaration
        if not is_base_and_derived(base, derived):
            return False
        for b in derived.recursive_bases:
            if ((b.related_class is base) and b.access_type !=
                    class_declaration.ACCESS_TYPES.PRIVATE):
                break
        else:
            return False

        base = target
        derived = source
        # d => b
        if self._is_both_declarated(base, derived):
            return True
        # d* => b*
        if type_traits.is_pointer(derived) and \
                type_traits.is_pointer(base) and \
                self._is_both_declarated(base.base, derived.base):
            return True
        # const d* => const b*
        if type_traits.is_pointer(derived) and \
                type_traits.is_pointer(base) and \
                type_traits.is_const(derived.base) and \
                type_traits.is_const(base.base) \
           and self._is_both_declarated(base.base.base, derived.base.base):
            return True
        # d* => const b*
        if type_traits.is_pointer(derived) and type_traits.is_pointer(base) \
           and type_traits.is_const(derived.base)\
           and self._is_both_declarated(base.base.base, derived.base):
            return True

        # d& => b&
        if type_traits.is_reference(derived) and \
                type_traits.is_reference(base) and \
                self._is_both_declarated(base.base, derived.base):
            return True
        # const d& => const b&
        if type_traits.is_reference(derived) and \
                type_traits.is_reference(base) and \
                type_traits.is_const(derived.base) and \
                type_traits.is_const(base.base) \
           and self._is_both_declarated(base.base.base, derived.base.base):
            return True
        # d& => const b&
        if type_traits.is_reference(derived) and \
                type_traits.is_reference(base) and \
                type_traits.is_const(derived.base) \
           and self._is_both_declarated(base.base.base, derived.base):
            return True
        return False

    def is_convertible(self):
        source = self.__source
        target = self.__target

        if self.__test_trivial(source, target):
            return True
        if type_traits.is_array(source) or type_traits.is_array(target):
            return False
        if self.__test_const_x_ref__to__x(source, target):
            return True
        if self.__test_const_ref_x__to__y(source, target):
            return True
        if self.__test_ref_x__to__x(source, target):
            return True
        if self.__test_ref_x__to__y(source, target):
            return True
        if self.__test_fundamental__to__fundamental(source, target):
            return True
        if self.__test_pointer_to_func_or_mv__to__func_or_mv(source, target):
            return True
        if self.__test_derived_to_based(source, target):
            return True

        if isinstance(source, cpptypes.declarated_t):
            if isinstance(source.declaration, enumeration.enumeration_t) \
               and type_traits.is_fundamental(target) \
               and not type_traits.is_void(target):
                return True  # enum could be converted to any integral type

            if isinstance(source.declaration, class_declaration.class_t):
                source_inst = source.declaration
                # class instance could be convertible to something else if it
                # has operator
                casting_operators = scopedef.find_all_declarations(
                    source_inst.declarations,
                    decl_type=calldef_members.casting_operator_t,
                    recursive=False)
                if casting_operators:
                    for operator in casting_operators:
                        if is_convertible(operator.return_type, target):
                            return True

        # may be target is class too, so in this case we should check whether
        # is has constructor from source
        if isinstance(target, cpptypes.declarated_t) and \
                isinstance(target.declaration, class_declaration.class_t):
            constructors = scopedef.find_all_declarations(
                target.declaration.declarations,
                decl_type=calldef_members.constructor_t,
                recursive=False)
            if constructors:
                for constructor in constructors:
                    if len(constructor.arguments) != 1:
                        continue
                    # TODO: add test to check explicitness
                    if is_convertible(source,
                                      constructor.arguments[0].decl_type):
                        return True

        return False


def is_convertible(source, target):
    """returns True, if source could be converted to target, otherwise False"""
    return __is_convertible_t(source, target).is_convertible()


def __is_noncopyable_single(class_, already_visited_cls_vars=None):
    """
    Implementation detail.

    Checks if the class is non copyable, without considering the base classes.

    Args:
        class_ (declarations.class_t): the class to be checked
        already_visited_cls_vars (list): optional list of vars that should not
            be checked a second time, to prevent infinite recursions.

    Returns:
        bool: if the class is non copyable
    """
    # It is not enough to check base classes, we should also to check
    # member variables.
    logger = utils.loggers.cxx_parser

    if has_copy_constructor(class_) \
       and has_public_constructor(class_) \
       and has_public_assign(class_) \
       and has_public_destructor(class_):
        msg = os.linesep.join([
            "__is_noncopyable_single - %s - COPYABLE:" % class_.decl_string,
            "    trivial copy constructor: yes",
            "    public constructor: yes",
            "    public assign: yes",
            "    public destructor: yes"])
        logger.debug(msg)
        return False

    if already_visited_cls_vars is None:
        already_visited_cls_vars = []

    if find_noncopyable_vars(class_, already_visited_cls_vars):
        logger.debug(
            ("__is_noncopyable_single(TRUE) - %s - contains noncopyable " +
             "members"), class_.decl_string)
        return True

    logger.debug((
        "__is_noncopyable_single(FALSE) - %s - COPYABLE, because is " +
        "doesn't contains noncopyable members"), class_.decl_string)
    return False


def is_noncopyable(class_, already_visited_cls_vars=None):
    """
    Checks if class is non copyable

    Args:
        class_ (declarations.class_t): the class to be checked
        already_visited_cls_vars (list): optional list of vars that should not
            be checked a second time, to prevent infinite recursions.
            In general you can ignore this argument, it is mainly used during
            recursive calls of is_noncopyable() done by pygccxml.

    Returns:
        bool: if the class is non copyable
    """
    logger = utils.loggers.cxx_parser

    class_decl = class_traits.get_declaration(class_)

    true_header = "is_noncopyable(TRUE) - %s - " % class_.decl_string

    if is_union(class_):
        return False

    if class_decl.is_abstract:
        logger.debug(true_header + "abstract client")
        return True

    # if class has public, user defined copy constructor, than this class is
    # copyable
    copy_ = find_copy_constructor(class_decl)
    if copy_ and copy_.access_type == 'public' and not copy_.is_artificial:
        return False

    if already_visited_cls_vars is None:
        already_visited_cls_vars = []

    for base_desc in class_decl.recursive_bases:
        assert isinstance(base_desc, class_declaration.hierarchy_info_t)

        if base_desc.related_class.decl_string in \
                ('::boost::noncopyable', '::boost::noncopyable_::noncopyable'):
            logger.debug(true_header + "derives from boost::noncopyable")
            return True

        if not has_copy_constructor(base_desc.related_class):

            base_copy_ = find_copy_constructor(base_desc.related_class)

            if base_copy_ and base_copy_.access_type == 'private':
                logger.debug(
                    true_header +
                    "there is private copy constructor")
                return True
            elif __is_noncopyable_single(
                    base_desc.related_class, already_visited_cls_vars):
                logger.debug(
                    true_header +
                    "__is_noncopyable_single returned True")
                return True

        if __is_noncopyable_single(
                base_desc.related_class, already_visited_cls_vars):
            logger.debug(
                true_header +
                "__is_noncopyable_single returned True")
            return True

    if not has_copy_constructor(class_decl):
        logger.debug(true_header + "does not have trivial copy constructor")
        return True
    elif not has_public_constructor(class_decl):
        logger.debug(true_header + "does not have a public constructor")
        return True
    elif has_destructor(class_decl) and not has_public_destructor(class_decl):
        logger.debug(true_header + "has private destructor")
        return True

    return __is_noncopyable_single(class_decl, already_visited_cls_vars)


def is_unary_operator(oper):
    """returns True, if operator is unary operator, otherwise False"""
    # definition:
    # member in class
    # ret-type operator symbol()
    # ret-type operator [++ --](int)
    # globally
    # ret-type operator symbol( arg )
    # ret-type operator [++ --](X&, int)
    symbols = ['!', '&', '~', '*', '+', '++', '-', '--']
    if not isinstance(oper, calldef_members.operator_t):
        return False
    if oper.symbol not in symbols:
        return False
    if isinstance(oper, calldef_members.member_operator_t):
        if len(oper.arguments) == 0:
            return True
        elif oper.symbol in ['++', '--'] and \
                isinstance(oper.arguments[0].decl_type, cpptypes.int_t):
            return True
        return False

    if len(oper.arguments) == 1:
        return True
    elif oper.symbol in ['++', '--'] \
            and len(oper.arguments) == 2 \
            and isinstance(oper.arguments[1].decl_type, cpptypes.int_t):
        # may be I need to add additional check whether first argument is
        # reference or not?
        return True
    return False


def is_binary_operator(oper):
    """returns True, if operator is binary operator, otherwise False"""
    # definition:
    # member in class
    # ret-type operator symbol(arg)
    # globally
    # ret-type operator symbol( arg1, arg2 )
    symbols = [
        ',', '()', '[]', '!=', '%', '%=', '&', '&&', '&=', '*', '*=', '+',
        '+=', '-', '-=', '->', '->*', '/', '/=', '<', '<<', '<<=', '<=', '=',
        '==', '>', '>=', '>>', '>>=', '^', '^=', '|', '|=', '||']
    if not isinstance(oper, calldef_members.operator_t):
        return False
    if oper.symbol not in symbols:
        return False

    if isinstance(oper, calldef_members.member_operator_t):
        if len(oper.arguments) == 1:
            return True
        return False

    if len(oper.arguments) == 2:
        return True
    return False


def is_copy_constructor(constructor):
    """
    Check if the declaration is a copy constructor,

    Args:
        constructor (declarations.constructor_t): the constructor
            to be checked.

    Returns:
        bool: True if this is a copy constructor, False instead.

    """
    assert isinstance(constructor, calldef_members.constructor_t)
    args = constructor.arguments
    parent = constructor.parent

    # A copy constructor has only one argument
    if len(args) != 1:
        return False

    # We have only one argument, get it
    arg = args[0]

    if not isinstance(arg.decl_type, cpptypes.compound_t):
        # An argument of type declarated_t (a typedef) could be passed to
        # the constructor; and it could be a reference.
        # But in c++ you can NOT write :
        #    "typedef class MyClass { MyClass(const MyClass & arg) {} }"
        # If the argument is a typedef, this is not a copy constructor.
        # See the hierarchy of declarated_t and coumpound_t. They both
        # inherit from type_t but are not related so we can discriminate
        # between them.
        return False

    # The argument needs to be passed by reference in a copy constructor
    if not type_traits.is_reference(arg.decl_type):
        return False

    # The argument needs to be const for a copy constructor
    if not type_traits.is_const(arg.decl_type.base):
        return False

    un_aliased = type_traits.remove_alias(arg.decl_type.base)
    # un_aliased now refers to const_t instance
    if not isinstance(un_aliased.base, cpptypes.declarated_t):
        # We are looking for a declaration
        # If "class MyClass { MyClass(const int & arg) {} }" is used,
        # this is not copy constructor, so we return False here.
        # -> un_aliased.base == cpptypes.int_t (!= cpptypes.declarated_t)
        return False

    # Final check: compare the parent (the class declaration for example)
    # with the declaration of the type passed as argument.
    return id(un_aliased.base.declaration) == id(parent)


def is_trivial_constructor(constructor):
    """
    Check if the declaration is a trivial constructor.

    Args:
        constructor (declarations.constructor_t): the constructor
            to be checked.

    Returns:
        bool: True if this is a trivial constructor, False instead.

    """
    assert isinstance(constructor, calldef_members.constructor_t)
    return not bool(constructor.arguments)



# Copyright 2014-2017 Insight Software Consortium.
# Copyright 2004-2009 Roman Yakovenko.
# Distributed under the Boost Software License, Version 1.0.
# See http://www.boost.org/LICENSE_1_0.txt

import pytest
import platform

from . import autoconfig

from pygccxml import parser
from pygccxml import declarations


TEST_FILES = [
    'test_smart_pointer.hpp'
]


@pytest.fixture
def global_ns():
    COMPILATION_MODE = parser.COMPILATION_MODE.ALL_AT_ONCE
    config = autoconfig.cxx_parsers_cfg.config.clone()
    if platform.system() == "Darwin":
        config.cflags = "-std=c++11 -Dat_quick_exit=atexit -Dquick_exit=exit"
        # https://fr.mathworks.com/matlabcentral/answers/2013982-clibgen-generatelibrarydefinition-error-the-global-scope-has-no-quick_exit-on-mac-m2#answer_1439856
        # https://github.com/jetbrains/kotlin/commit/d50f585911dedec5723213da8835707ac95e1c01
    else:
        config.cflags = "-std=c++11"
    decls = parser.parse(TEST_FILES, config, COMPILATION_MODE)
    global_ns = declarations.get_global_namespace(decls)
    global_ns.init_optimizer()
    return global_ns


def test_is_smart_pointer(global_ns):
    """
    Test smart_pointer_traits.is_smart_pointer method.

    """

    criteria = declarations.declaration_matcher(name="yes1")
    decls = declarations.matcher.find(criteria, global_ns)
    assert declarations.smart_pointer_traits.is_smart_pointer(
        decls[0].decl_type) is True

    criteria = declarations.declaration_matcher(name="no1")
    decls = declarations.matcher.find(criteria, global_ns)
    assert declarations.smart_pointer_traits.is_smart_pointer(
            decls[0].decl_type) is False

    criteria = declarations.declaration_matcher(name="no2")
    decls = declarations.matcher.find(criteria, global_ns)
    assert declarations.smart_pointer_traits.is_smart_pointer(
            decls[0].decl_type) is False


def test_is_auto_pointer(global_ns):
    """
    Test auto_ptr_traits.is_smart_pointer method.

    """

    criteria = declarations.declaration_matcher(name="yes2")
    decls = declarations.matcher.find(criteria, global_ns)
    assert declarations.auto_ptr_traits.is_smart_pointer(
        decls[0].decl_type) is True

    criteria = declarations.declaration_matcher(name="no1")
    decls = declarations.matcher.find(criteria, global_ns)
    assert declarations.auto_ptr_traits.is_smart_pointer(
        decls[0].decl_type) is False

    criteria = declarations.declaration_matcher(name="no2")
    decls = declarations.matcher.find(criteria, global_ns)
    assert declarations.auto_ptr_traits.is_smart_pointer(
        decls[0].decl_type) is False


def test_smart_pointer_value_type(global_ns):
    """
    Test smart_pointer_traits.value_type method.

    """

    criteria = declarations.declaration_matcher(name="yes1")
    decls = declarations.matcher.find(criteria, global_ns)
    vt = declarations.smart_pointer_traits.value_type(decls[0].decl_type)
    assert isinstance(vt, declarations.int_t) is True


def test_auto_pointer_value_type(global_ns):
    """
    Test auto_pointer_traits.value_type method.

    """

    criteria = declarations.declaration_matcher(name="yes2")
    decls = declarations.matcher.find(criteria, global_ns)
    vt = declarations.auto_ptr_traits.value_type(decls[0].decl_type)
    assert isinstance(vt, declarations.double_t) is True
